/*  NANO Éphéméride : Programme noyau pour tester rapidement des fonctions.
     
DIFFERENTS BRANCHEMENTS :
 D0  : Libre pour F.E.                         A0 : Lecture du clavier à deux B.P. FC- et FC +.
 D1  : Libre pour F.E.                         A1 : Lecture du bouton central du capteur rotatif.
 D2  : Codeur Incrémental B.                   A2 : Libre pour F.E.
 D3  : Codeur Incrémental A.                   A3 : Libre pour F.E.    
 D4  : Libre pour F.E.                         A4 : Libre pour F.E.
 D5  : Libre pour F.E.                         A5 : Libre pour F.E.
 D6  : Libre pour F.E. 
 D7  : Libre pour F.E.                         A6 : Libre pour F.E.
 D8  : LED jaune d'avertissements divers.      A7 : Libre pour F.E.
 D9  : Libre pour F.E.
 D10 : Pilotage du BUZZER.
 D11 : Broche_SDA Pilotage OLED.
 D12 : Broche_SCL Pilotage OLED.
 D13 : LED Verte d'Arduino.        
 */

#include <stdio.h>  // Impératif pour l'HORLOGE DS1302.
#include <DS1302.h> // Bibliothèque pour l'HORLOGE DS1302.

// Dédinir les E/S sur Arduino.
uint8_t LigneRST  = 4;
uint8_t LigneIO   = 5;
uint8_t LigneSCLK = 6;

// Créer des "registres tampon".
char TamponDeDonnees[50]; char JOUR[10];

// Creer "l'objet" DS1302.
DS1302 rtc(LigneRST, LigneIO, LigneSCLK);

#include <Adafruit_ssd1306syp.h> // Impérativement avant #include <EEPROM.h>.

// Déclarations pour le LOGO.
#include <EEPROM.h>
#define XL 5  // Position du LOGO.
#define YL 37 // Position du LOGO.

// Déclarations pour l'afficheur OLED :
#define broche_SDA 11 // D11 utilisée en sortie.
#define broche_SCL 12 // D12 utilisée en sortie.
Adafruit_ssd1306syp display(broche_SDA, broche_SCL);

// Déclarations pour le capteur codeur rotatif KY-40 :
#define Codeur_A 3   // D3 : INT 1 sur broche 5 du microcontrôleur. (CLK)
#define Codeur_B 2    // D2 : INT 0 sur broche 4 du microcontrôleur. (DT)
#define BP_central 15 // A1 pour lire le bouton central.

// Constantes du programme.
#define LED_Jaune 8  // LED d'avertissements divers.
#define BUZZER  10   // Sortie dédiée au BUZZER.
#define LED_Arduino 13
#define NB_tests_antirebonds 50 // Nombre de tests de stabilisation B.P.

// Variables du programme.

boolean Clic_sur_le_BP, Augmenter_le_choix, Clic_long;
volatile boolean Rotation;  // Indique une activation du codeur rotatif.
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif;   // Sens de rotation.

void setup() {
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  pinMode(LED_Arduino, OUTPUT);  digitalWrite(LED_Arduino, LOW);
  pinMode(LED_Jaune, OUTPUT); digitalWrite(LED_Jaune, LOW); pinMode(BUZZER, INPUT);
  digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A1.
  display.initialize(); delay(300);
  display.setTextSize(1); // Taille la plus petite.
  }
  
void loop() {
  AfficheLeTemps(); delay(1000);
  //======================== Traite le B.P. central ======================
  //========================= Traite la rotation =========================
  INT_active = true; // Armer un nouveau comptage.
  if (Rotation) { }
  //=============== Traiter la requête de l'un des deux B.P. =============
  Tester_le_BP_A0();
  if (Clic_sur_le_BP) {
      Clic_sur_le_BP = false;} // Action sur le B.P. prise en compte.
}//============================= Fin de la boucle de base ================

void AfficheLeTemps() { // Télécharger la date et l'heure préservée dans le DS1302.
  Time t = rtc.time();
  // Nommer les jours de la semaine.
  display.clear(); display.setCursor(10,30);
  display.print(t.date); display.print(" "); display.print(t.mon); display.print(" "); display.print(t.yr);
  display.setCursor(10,50);
  display.print(t.hr); display.print(" "); display.print(t.min); display.print(" "); display.print(t.sec);
  display.update();
}

//-------------------- Routines de servitude pour traiter le capteur incrémental -------------------
void Delay_antirebonds() {for(int I=0; I<3; I++);} // Valeur de 3 expérimentale.

void Front_sur_A(){ // Interruption quand A change d'état avant B.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
  if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
    A_set = !A_set;
    if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }

void Front_sur_B(){ // Interruption quand B change d'état avant A.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
  if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
    B_set = !B_set;
    // Décrémenter le compteur si B avant A.
    if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
//--------------------------------------------------------------------------------------------------

void Effacer_zone(byte X, byte Y, byte NB_caracteres) {
  byte I = 10; 
  for (byte J=0; J < I; J++)
       display.drawLine(X, Y + J, X + (NB_caracteres * 6), Y + J, BLACK);}
  
void Affiche_cadre_du_haut(byte X) {
  CADRE_UP();  display.setTextSize(1); // Taille la plus petite.
  display.setCursor(X,4);}

void CADRE_UP() {display.drawRect(0, 0, 128, 16, WHITE);}
void CADRE() {display.drawRect(0, 16, 128, 48, WHITE);}

void BIP() {
  pinMode(BUZZER, OUTPUT); tone(BUZZER,3000); delay(100); noTone(BUZZER); pinMode(BUZZER, INPUT);}
  
void ATTENDRE_un_BP() { Clic_sur_le_BP = false;
  while (!Clic_sur_le_BP) {
    digitalWrite(LED_Arduino,!digitalRead(LED_Arduino)); // Inverser la LED clavier.
    delay(50); Tester_le_BP_A0();}
  digitalWrite(LED_Arduino,LOW); Clic_sur_le_BP = false; }
  
// >>>>>>>>>>>>>>> Le B.P. est câblé en "Analogique" <<<<<<<<<<<<<<<<<<
void Tester_le_BP_A0() { // La lecture sur A0 donne 0 ou 930 maximum.
  int Numerisation_BP; // Valeur retournée par la CAN de A0.
  boolean Sauve_etat_LED_jaune;
  byte COMPTEUR; unsigned long Horloge_actuelle;
  Sauve_etat_LED_jaune = digitalRead(LED_Jaune); Numerisation_BP = analogRead(0);
  if ((Numerisation_BP > 710) || (Numerisation_BP < 240)) {
    Clic_sur_le_BP = true; digitalWrite(LED_Arduino, HIGH);
    Augmenter_le_choix = false;
    Horloge_actuelle = millis(); Clic_long = false;
    if (Numerisation_BP > 710) { Augmenter_le_choix = true; }
    delay(100); // Attendre_la_stabilisation du travail.
    // Attendre_la_stabilisation du repos :
    COMPTEUR = 0;
    while (COMPTEUR < NB_tests_antirebonds) {
      if ((millis() - Horloge_actuelle) > 700) {Clic_long = true; digitalWrite(LED_Jaune, HIGH);}
      COMPTEUR++; Numerisation_BP = analogRead(0);
      if ((Numerisation_BP > 710) || (Numerisation_BP < 240)) COMPTEUR = 0; }
  digitalWrite(LED_Arduino, LOW);} // Éteindre la LED clavier.
  digitalWrite(LED_Jaune, Sauve_etat_LED_jaune); } // Restitue l'état de la LED jaune.
  
