
/*  NANO Éphéméride Solaire avec affichage sur un module OLED SSD1306.
           Par Michel DROUI. (Pseudo Internet : Nulentout.)
           
    Version la plus simple. Actuellement 27352 Octets (89%) et 356 Octets (17%). (Avec bibliothèque.)
    
    La latitude est comprise entre 40° et 52°.
    Les longitudes possibles sont imitées entre -8° et +8°.
    L'ALTITUDE est limitée entre zéro et 9999m.
    
    Détermine le changement d'heure avec une imprecision annuelle de + ou - 3 Jours maximum.
           
    Le B.P. central est affecté au MODE GRAPHIQUE.
    Le bouton rotatif hors mode graphique fait changer de mode écran :
          0 : Mode veille > Écran noir.
          1 : Affichage de la position du Soleil.
          2 : Affichage du Lever et du Coucher.
          3 : Affichage de l'Équation du temps.
          4 : Affichage d'informations complémentaires : Jour Julien, durée de la nuit ...
          5 : Affichage du mode crépusculaire.
          6 : Affichage de la situation géographique.
    FC- court : Changer le Mode crépusculaire.
    FC+ court : Changer la DATE d'Observation.
                FC+ court change de donnée.
                FC- court recircule des dates particulières.         
    FC- long : Changer le LIEU de l'observatoire. 
    FC+ long : Changer l'HEURE d'Observation.        

DIFFERENTS BRANCHEMENTS :
 D0  : Libre pour F.E.                         A0 : Lecture du clavier à deux B.P. FC- et FC +.
 D1  : Libre pour F.E.                         A1 : Lecture du bouton central du capteur rotatif.
 D2  : Codeur Incrémental B.                   A2 : Libre pour F.E.
 D3  : Codeur Incrémental A.                   A3 : Libre pour F.E.    
 D4  : Libre pour F.E.                         A4 : Libre pour F.E.
 D5  : Libre pour F.E.                         A5 : Libre pour F.E.
 D6  : Libre pour F.E. 
 D7  : Libre pour F.E.                         A6 : Libre pour F.E.
 D8  : LED jaune d'avertissements divers.      A7 : Libre pour F.E.
 D9  : Libre pour F.E.
 D10 : Pilotage du BUZZER.
 D11 : Broche_SDA Pilotage OLED.
 D12 : Broche_SCL Pilotage OLED.
 D13 : LED Verte d'Arduino.        

// La SALAMANDRE (LOGO) a été écrite en EEPROM avec P1.
// Les textes en EEPROM sont écrits avec P2.
// Avant de passer les textes en EEPROM : 23484 (76%) 799 (39%). (Collision de PILE constatée.)
// Avant de passer les textes en EEPROM : 23944 (76%) 315 (15%). (Pb de collision éliminé.)           */

#include <Adafruit_ssd1306syp.h> // Impérativement avant #include <EEPROM.h>.

// Déclarations pour le LOGO.
#include <EEPROM.h>
#define XL 8  // Position du LOGO.
#define YL 34 // Position du LOGO.

// Déclarations pour l'afficheur OLED :
#define broche_SDA 11 // D11 utilisée en sortie.
#define broche_SCL 12 // D12 utilisée en sortie.
Adafruit_ssd1306syp display(broche_SDA, broche_SCL);

// Déclarations pour le capteur codeur rotatif KY-40 :
#define Codeur_A 3   // D3 : INT 1 sur broche 5 du microcontrôleur. (CLK)
#define Codeur_B 2    // D2 : INT 0 sur broche 4 du microcontrôleur. (DT)
#define BP_central 15 // A1 pour lire le bouton central.

// Constantes du programme.
#define Version_PGM "Version 24-04-17"
#define Max_mode_ecran 6 // Nombre de modes écran - 1.
#define LED_Jaune 8  // LED d'avertissements divers.
#define BUZZER  10   // Sortie dédiée au BUZZER.
#define LED_Arduino 13
#define NB_tests_antirebonds 50 // Nombre de tests de stabilisation B.P.

// Variables du programme.

boolean Clic_sur_le_BP, Augmenter_le_choix, Clic_long;
volatile boolean Rotation;  // Indique une activation du codeur rotatif.
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif;   // Sens de rotation.
boolean Edite_HEURE;    // On édite soit l'HEURE, soit la DATE.
boolean Aff_graduations; // Affiche les graduations en mode graphique.
boolean Mode_trois_courbes, Edit_Longitude, Affiche_date;
boolean Longitude_EST;
byte OCTET, Ligne; // Pour le LOGO.
byte X; // Position sur X pour les titres et pour le LOGO.
byte Y; // Position sur Y pour tracer les pixels.
byte PTR; // Pour le LOGO et le pointeur en RAM d'échantillons.
byte Mode_Ecran = 0;
byte Coef_deplacement, Mode_Edition;
byte Mode_Crepuscule = 0;
	//   -  0 : Crépuscule sur l'horizon avec correction de la réfraction athmosphérique soit 0.833°
	//   -  6 : Crépuscule civil
	//   -  12 : Crépuscule nautique
	//   -  18 : Crépuscule astronomique
byte Mois = 1; byte Jour = 1; byte Heure = 12; byte Minutes = 00; byte Secondes = 00;
byte SAVM = 1, SAVJ = 1; 
byte savH, savM, savS, savJ, savMois; 
int Annee = 2017;
int Pos_X_sur_ecran;
int Num_Jour_dans_annee;
int Colonne; // Int car utilisé pour "Jour" allant jusqu'à 365;
float VALEUR;

//***************************** Déclarations spécifiques aux calcules astronomiques ***************************
#define Nombre_de_degres_par_Radian 57.295779505601 // Nombre de degrés par Radian.
double Latitude = 44.4491;  // (Arcambal)
double Longitude = 1.43663;  // (Arcambal)
int Altitude = 142; // Altitude d'observation. (142 pour Arcambal)
boolean Heure_Hiver;
boolean Bissextile;
float Equation_du_Temps; // Float car peut être négatif et décimal.
int nbJ_par_mois[] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
float declinaison; // Déclinaison en radians.
float Hauteur;
float Azimut;
String Lever, Culmine, Coucher;
float Lever_en_min, Coucher_en_min;
//*************************************************************************************************************

void setup() {
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  pinMode(LED_Arduino, OUTPUT);  digitalWrite(LED_Arduino, LOW);
  pinMode(LED_Jaune, OUTPUT); digitalWrite(LED_Jaune, LOW); pinMode(BUZZER, INPUT);
  digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A1.
  display.initialize(); delay(300); display.clear();
  display.setTextSize(1); // Taille la plus petite.
  display.setCursor(18,4);
  display.print(Version_PGM); 
  CADRE(); display.setTextSize(2); // Taille double.
  display.setCursor(47,21); Aff_TEXTE_EEPROM(120, 6); 
  //================================= Tracer le LOGO. ===============================
  PTR = 0; // Pointer le premier octer dans l'EEPROM.
  for(Ligne = 0; Ligne < 24; Ligne++)
     {for(Colonne = 0; Colonne < 5; Colonne++) 
         {OCTET = EEPROM.read(PTR);
          for(byte I=0; I < 8; I++)
              {display.drawPixel((Colonne*8)+XL+I, Ligne+YL, (OCTET & B10000000));
               OCTET=OCTET<<1;}; PTR++; } };
  display.drawLine(XL+40, YL+18, XL+40, YL+22,WHITE); display.drawLine(XL+41, YL+19, XL+41, YL+21,WHITE); 
  //=================================================================================
  display.setTextSize(1); display.setCursor(70,45); Aff_TEXTE_EEPROM(127, 5); 
  display.update();
  ATTENDRE_un_BP(); // Attend un BP pour sortir. // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  Rotation = true; Sens_positif = true; // Afficher un premier écran au démarrage.
  Longitude_EST = true; Effectuer_les_calculs();} // Avoir des valeurs cohérentes au démarrage.
  
void loop() {
  //======================== Traite le B.P. central ======================
  Traite_BP_Central();
  //========================= Traite la rotation =========================
  INT_active = true; // Armer un nouveau comptage.
  if (Rotation) { Rotation = false;
     if (Sens_positif) {Mode_Ecran++; if (Mode_Ecran > Max_mode_ecran) Mode_Ecran = 0;}
     else {if (Mode_Ecran==0) Mode_Ecran = Max_mode_ecran;  else Mode_Ecran--;}
     Effectuer_les_calculs();
     switch (Mode_Ecran) {
        case 0 : {Efface_ecran(); break;} // Mode veille > Écran noir.
        case 1 : {Affiche_Position_Soleil(); break;} // Affichage de la position du Soleil.
        case 2 : {Affiche_Lever_Coucher(); break;} // Affichage du Lever et du Coucher.
        case 3 : {Coef_deplacement = 1; Trace_Equation_du_temps(); break;} // Affichage l'équation du temps.
        case 4 : {Affichages_complementaires(); break;} // Durée du jour, heure Été/Hiver ...
        case 5 : {Affiche_mode_crepusculaire(); break;} // Affichage du mode crépusculaire.
        case 6 : Affiche_la_Position();} } // Affichage de la situation géographique. 
//=============== Traiter la requête de l'un des deux B.P. =============
  Tester_le_BP_A0();
  if (Clic_sur_le_BP) {Clic_sur_le_BP = false; // Action sur le B.P. prise en compte.
      if (Mode_Ecran == 3)
      //==================== Écran sur affichage Équation du temps ====================
         //================ Ci-dessous traitement du clic long =================
         if (Clic_long) {
             if (Augmenter_le_choix) {
                 switch (Coef_deplacement) {
                     case 30 : {Coef_deplacement = 1; break;}
                     case 15 : {Coef_deplacement = 30; break;}
                     case 7 : {Coef_deplacement = 15; break;}
                     case 4 : {Coef_deplacement = 7; break;}
                     case 2 : {Coef_deplacement = 4; break;}
                     case 1 : Coef_deplacement = 2;}
                 Afficher_coef_deplacement();}
             else {Affiche_date = true; Num_Jour_dans_annee = Jour_Num(Annee, Mois, Jour);
                  Afficher_Jour_valeurET_Curseur_Bissextile();}}
         //================ Ci-dessous traitement du clic court ================
         else {if (Augmenter_le_choix) 
                  {Num_Jour_dans_annee = Num_Jour_dans_annee + Coef_deplacement;
                   if (Num_Jour_dans_annee > 365) Num_Jour_dans_annee = 1;}
               else {if (Num_Jour_dans_annee >= Coef_deplacement)
                         Num_Jour_dans_annee = Num_Jour_dans_annee - Coef_deplacement;
                     else Num_Jour_dans_annee = 365;}
               Affiche_date = false; Afficher_Jour_valeurET_Curseur_Bissextile();}                     
      //======================= Écrans ordinaires du MENU de base =====================
      else   if (Clic_long) if (Augmenter_le_choix) Changer_Heure_Observation();
                            else Changer_le_lieu_Observation();
             else if (Augmenter_le_choix) Changer_Date_Observation();
                  else Changer_mode_crepusculaire();}
}//============================= Fin de la boucle de base ===================================

void ETE_HIVER() {
  // Détermine le changement d'heure avec une imprecision annuelle de + ou - 3 Jours maximum.
  byte NumJ = Jour_Num(Annee, Mois, Jour);
  Heure_Hiver = true;
  if((NumJ > 88) && ((NumJ < 331))) Heure_Hiver = false;}
  
void Affichages_complementaires() {
  display.clear(); Affiche_cadre_du_haut(12);
  Aff_TEXTE_EEPROM(813, 7); Affiche_la_date(); ETE_HIVER(); 
  CADRE(); display.setCursor(23,20); Aff_TEXTE_EEPROM(628, 8);
  if (Heure_Hiver) Aff_TEXTE_EEPROM(834, 6); else {display.print("ete."); Accent_aigu(72,20); Accent_aigu(84,20);}
  display.setCursor(118,20); if (Bissextile) display.print("B");
                        else {display.drawLine(114, 17, 126, 29, WHITE); display.drawLine(114, 29, 126, 17, WHITE);}
  display.drawLine(113, 17, 113, 29, WHITE); 
  display.drawLine(1, 30, 126, 30, WHITE); display.setCursor(10,33); Aff_TEXTE_EEPROM(841, 11);
  // Début calendrier Julien : Premier Janvier 4712 avant J.C.
  display.print(long((4712 + Annee) * 365.24806) + Jour_Num(Annee, Mois, Jour));
  display.setCursor(10,43); Aff_TEXTE_EEPROM(820, 7);
  display.print(Conv_Temps_en_chaine(Coucher_en_min - Lever_en_min - (2 * abs(Equation_du_Temps))));
  display.setCursor(10,53); Aff_TEXTE_EEPROM(827, 7);
  display.print(Conv_Temps_en_chaine((1440 - Coucher_en_min) + Lever_en_min + (2 * abs(Equation_du_Temps))));
  display.update();}

//******************************** Routines spécifiques aux calculs astronomiques *****************************
int Jour_Num(int A, int M, int J) { // Retourne le jour de l'année pour la Date définie.
  int nbJ = 0; for (byte i = 0; i < M; i++) nbJ += nbJ_par_mois[i]; if (Bissextile) nbJ++;  return nbJ + J;}
  
void Calcule_Equation_du_temps(int Ordre_du_jour) {// Valeur retournée en minutes et fractions de minutes.
  float Angle_orbital = 6.283185308 / 365.25 * Ordre_du_jour; 
  Equation_du_Temps = (7.66 * sin(Angle_orbital)) + (9.87 * sin((2 * (Angle_orbital+0.174533))));}
  
float Angle_orbital_Date_heure(int A, int M, int J, int H, int Mn, int S) {
  // Fraction de l'année en radians le jour de l'année "a", au mois "m", au jour "j", 
  // à l'heure "h", la minute "mn" et la seconde "s"
  int Nb = 365; if (Bissextile) Nb = 366;
  return 6.283185308 * (Jour_Num(A, M, J) - 1 + (H + Mn / 60 + S / 3600 - 12) / 24) / Nb;} // Radians

void eqT_Decl(float Angle_orbital) {
  // Retourne l'équation du temps en minutes et l'angle de déclinaison du soleil en radians
  Equation_du_Temps = (7.66 * sin(Angle_orbital)) + (9.87 * sin((2 * (Angle_orbital + 0.174533))));
  declinaison = 0.006918 - 0.399912 * cos(Angle_orbital) + 0.070257 * sin(Angle_orbital)
                - 0.006758 * cos(2 * Angle_orbital) + 0.000907 * sin(2 * Angle_orbital)
                - 0.002697 * cos(3 * Angle_orbital) + 0.00148 * sin(3 * Angle_orbital);} // Radians
                
void posSoleil(int A, int M, int J, int H, int Mn, int S, float latitude,
               float longitude, float *Hauteur, float *Azimut) {
  // Latitude et longitude du lieu exprimées en degrés décimaux.
  // La latitude est comprise entre +40° et +52°. (Hémisphère Nord.) (Donc positive dans les calculs.)
  // La longitude est comprise entre -9° et +9°. (Positive si EST et négative si OUest.)
  // L'azimut du soleil est compté positivement dans le sens horairee à partir du Nord.
  eqT_Decl(Angle_orbital_Date_heure(A, M, J, H, Mn, S));
  // Heure légale = Heure solaire vraie + Equation_du_Temps - (4 min par ° EST) + (1H ou 2H si Été ou Hiver) :
  float Correction_pour_Heure_legale = +Equation_du_Temps + (4.0 * longitude) - 120 + (60 * Heure_Hiver);
  float HeureSolaireVraie = H * 60.0 + Mn + (S / 60.0) + Correction_pour_Heure_legale; // En minutes.
  //======================== Ci-dessous : Angle_Solaire_Relatif exprimé en degrés. ========================
  float aSr = (HeureSolaireVraie / 4.0) - 180;
  // Si HSV = 0H aSr = -180 / Si HSV = 12H aSr = 0 / Si HSV = 18H aSr = 90 / Si HSV = 24H aSr = 180 
  latitude = radians(latitude); longitude = radians (longitude);
  float ze = acos(sin(latitude) * sin(declinaison)
             + cos(latitude) * cos(declinaison) * cos(radians(aSr))); // En radians.
  // Ci-dessous, cs est le cosinus de l'azimuth exprimé en radians.
  float cs = -(sin(latitude) * cos(ze) - sin(declinaison)) / (cos(latitude) * sin(ze)); // En radians.
  if (cs < -1)  cs = -1;   // Correction pour palier les imprécisions de calcul.
  else if (cs > 1) cs = 1; // Correction pour palier les imprécisions de calcul.
  *Azimut = degrees(acos(cs)); // Conversion du cosinus  en angle, exprimé en degrés décimaux.
  //================================== Ci-dessous : Calcul de la hauteur ==================================
  if (aSr > 0) *Azimut = 360 - *Azimut; *Hauteur = 90 - (degrees(ze));} // En degrés décimaux.
  
String Conv_Temps_en_chaine(float Temps_en_minutes) {
  // Retourne une chaine au format HH:MM:SS
  // L'entrée est le temps exprimé en minutes et dixièmes de minutes.
  int Hr = Temps_en_minutes / 60;
  int Min = trunc(Temps_en_minutes - 60 * Hr);
  int Sec = trunc((Temps_en_minutes - trunc(Temps_en_minutes)) * 60);
  char stx[10]; // stockage de la chaîne complète
  sprintf(stx, "%02u:%02u:%02u", Hr, Min, Sec); return stx;}  
  
void lmvSoleil(int a, int m, int j, byte Type_de_crepuscule,
               float latitude, float longitude, String *Lever, String *Culmine, String *Coucher) {
  // "EteHiv" vaut 60 en hiver et 120 en Été. (décalage horaire exprimé en minutes.)
  // "Type_de_crepuscule" est le type d'aurore ou de crépuscule retenu :
  //       0 : Crépuscule sur l'horizon avec correction de la réfraction athmosphérique  : r = 0.833°
  //       6 : Crépuscule civil. 12 : Crépuscule nautique. 18 : Crépuscule astronomique.
  eqT_Decl(Angle_orbital_Date_heure(a, m, j, 12, 0, 0));
  latitude = radians(latitude);
  byte EteHiv = 120; if (Heure_Hiver) EteHiv = 60;
  float Acr; // Cosinus de l'Angle crépusculaire.
  switch (Type_de_crepuscule) {
    case 0 : {Acr = cos(radians((90.833 + (0.015 * sqrt(Altitude))))); break;}  // Horizon + correction réfraction.
    case 6 : {Acr = cos(radians(96)); break;} // Aurore et Crépuscules Civils.
    case 12 : {Acr = cos(radians(102)); break;} // Aurore et Crépuscules Nautiques.
    case 18 : Acr = cos(radians(108)); } // Aurore et Crépuscules Astronomiques.
  float aS = acos(Acr / (cos(latitude) * cos(declinaison)) - tan(latitude) * tan(declinaison))
             * Nombre_de_degres_par_Radian;
  Lever_en_min = 720 + (4 * (-longitude - aS)) - Equation_du_Temps + EteHiv;
  Coucher_en_min = 720 + (4 * (-longitude + aS)) - Equation_du_Temps + EteHiv;
  *Lever   = Conv_Temps_en_chaine(Lever_en_min);
  *Coucher = Conv_Temps_en_chaine(Coucher_en_min);
  *Culmine = Conv_Temps_en_chaine(720 + (4 * longitude) - Equation_du_Temps + EteHiv);}
//*************************************************************************************************************

void Afficher_coef_deplacement() {
   Effacer_zone(87, 45, 2); display.setCursor(87,45);
   if(Coef_deplacement < 10) display.print("x"); display.print(Coef_deplacement); display.update();}
                 
void Afficher_Jour_valeurET_Curseur_Bissextile() {
  // Toutes ces données sont affichées en fonction de la variable [Num_Jour_dans_annee].
  Effacer_les_donnees(); display.setCursor(13,4); // Pour afficher le texte dans le cadre jaune.
  if (Affiche_date) {Aff_TEXTE_EEPROM(620, 8); Affiche_la_date();} else Aff_TEXTE_EEPROM(794, 18);
  Bissextile = (Annee % 4 == 0); if (Bissextile) {display.setCursor(40,24); display.print("B");
  display.drawRect(38, 22, 9, 12, WHITE);} else {display.drawRect(38, 22, 9, 12, BLACK);
  Effacer_zone(39, 23, 1);} display.setCursor(89,24); display.print(Num_Jour_dans_annee);
  Calcule_Equation_du_temps(Num_Jour_dans_annee); display.setCursor(40, 51);
  display.print(Equation_du_Temps,1); Colonne = map (Num_Jour_dans_annee,1,365,1,110); Affiche_le_curseur();
  display.update();}

void Efface_le_curseur() {
     display.drawLine(1, 17, 110, 17, BLACK); display.drawLine(1, 18, 110, 18, BLACK);}
     
void Effacer_les_donnees() {
  Effacer_zone(13, 4, 18); // Efface le texte dans le cadre jaune.
  // Effacer les données J, E.T. et Curseur :
  Effacer_zone(89, 24, 3); Effacer_zone(40, 51, 5); Efface_le_curseur();}

void Trace_Equation_du_temps() {
  byte X, Y; display.clear(); Affiche_date = true;
  Rectangle_Graphe(); CADRE_UP();
  // Tracer les graduations sur le coté droit: 
  for (byte I = 19; I < 64; I=I+7) display.drawLine(111, I, 114, I, WHITE);
  // Tracer les valeurs latérales :
  display.setCursor(116,23); display.print("10"); display.setCursor(116,37); display.print("0");
  display.setCursor(116,51); display.print("10"); display.drawLine(1, 40, 110, 40, WHITE);
  display.setCursor(77,24); display.print("J="); display.setCursor(10,51); display.print("E.T.=");
  Afficher_coef_deplacement();
  //================== Affiche le graphe de l'équation du temps; ======================
  for (int J = 1; J < 366; J++) {
       Calcule_Equation_du_temps(J);
       X = map (J,1,365,1,110);
       Y = map (Equation_du_Temps,-17,+17,63,18);
       display.drawPixel(X, Y, WHITE);}
  //===================================================================================
  Num_Jour_dans_annee = Jour_Num(Annee, Mois, Jour); Afficher_Jour_valeurET_Curseur_Bissextile();  
  display.update();} 
  
void Rectangle_Graphe() {display.drawRect(0, 16, 112, 48, WHITE);}
  
void Effectuer_les_calculs() {
  double Longitude_avec_signe;
  Longitude_avec_signe = Longitude; if (!Longitude_EST) Longitude_avec_signe = -Longitude_avec_signe;
  ETE_HIVER(); // Mise à jour heure Été/Hiver.
  Bissextile = (Annee % 4 == 0); // Détermine si bissextile.
  posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude_avec_signe, &Hauteur, &Azimut);
  //-------------------------- Calcule Lever, Méridien, Coucher -------------------------
  lmvSoleil(Annee, Mois, Jour, Mode_Crepuscule, Latitude, Longitude_avec_signe, &Lever, &Culmine, &Coucher);}
  
void Faire_evoluer_le_temps() { // On calcule toutes les deux minutes.
  Minutes = Minutes + 2; if (Minutes > 59) {Minutes = 0; Heure++;}}
  
void Trace_Soleil() {
  byte X,Y;
  savH = Heure; savM = Minutes; savS = Secondes; Heure = 2; Minutes = 0; Secondes = 0;
  //================= Attendre d'être dans la zone affichable ================== 
  while (Azimut < 40) { 
         posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude, &Hauteur, &Azimut);
         Faire_evoluer_le_temps();}
         Faire_evoluer_le_temps();
         posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude, &Hauteur, &Azimut);
  while (Hauteur < -20) {
        Faire_evoluer_le_temps();
        posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude, &Hauteur, &Azimut);}
  //================ Tracer les points dans la zone affichable =================
  while ((Hauteur > -20) && (Azimut < 308))
        {Faire_evoluer_le_temps(); posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude, &Hauteur, &Azimut);
         X = map(Azimut,41,308,2,112); Y = map(Hauteur,-20,70,63,18);
         display.drawPixel(X, Y, WHITE);}
  //============================================================================
  Heure = savH; Minutes = savM; Secondes = savS;}

void Affiche_les_reperes(){
  for (byte I=4; I < 112; I=I+4) display.drawLine(I, 59, I, 62, WHITE); 
  display.drawLine(20, 17, 20, 56, WHITE); display.drawLine(56, 17, 56, 56, WHITE);  
  display.drawLine(92, 17, 92, 56, WHITE);
  Effacer_zone(17, 57, 1); display.setCursor(18,57); display.print("E");
  Effacer_zone(52,57, 1); display.setCursor(53,57); display.print("S");
  Effacer_zone(89, 57, 1); display.setCursor(90,57); display.print("W");
  // Tracer les repères horizontaux : 
  for (byte I=23; I < 50; I=I+10) 
       for (byte J=0; J < 109; J=J+5) display.drawLine(J+1, I, J+2, I, WHITE);}

void Efface_zone_date() {Effacer_zone(7,4,10); Effacer_zone(75,4,8);}
  
void Affiche_ecran_Graphique() {
  display.clear(); Mise_a_jour_Date_et_Heure(); Rectangle_Graphe();
  if (Aff_graduations) Affiche_les_reperes();
  // Tracer les graduations sur le coté droit: 
  for (byte I = 18; I < 64; I=I+5) display.drawLine(111, I, 114, I, WHITE);
  // Tracer les valeurs latérales.
  display.setCursor(116,20); display.print("60"); display.setCursor(116,30); display.print("40");  
  display.setCursor(116,40); display.print("20"); display.setCursor(116,50); display.print("0");
  display.drawLine(1, 53, 110, 53, WHITE);
  Efface_zone_date(); display.setCursor(7,4); Aff_TEXTE_EEPROM(133, 19); display.update();
  if (Mode_trois_courbes)
     {savJ = Jour; savMois = Mois; Mois = 6; Jour = 21; Trace_Soleil(); Mois = 9; Jour = 23; Trace_Soleil();
      Mois = 12; Jour = 21; Trace_Soleil(); Jour = savJ; Mois = savMois;}
  else Trace_Soleil();
  if (Mode_trois_courbes)
       {Effacer_zone(7,4,19); display.drawLine(70, 1, 70, 14, BLACK);
        display.setCursor(8,4); Aff_TEXTE_EEPROM(153, 19);}
  else {Efface_zone_date(); Mise_a_jour_Date_et_Heure();}
  display.update();}
  
void Affiche_le_curseur() {
  display.drawLine(Colonne-1, 17, Colonne+1, 17, WHITE);
  display.drawPixel(Colonne, 18, WHITE); display.update();}
  
void Affiche_coef_deplacement() {
  Effacer_zone(96, 20, 2); display.setCursor(96,20); display.print("x");
  display.print(OCTET); display.update();}
  
void Determine_parametres_pour_le_curseur() {
  savH = Heure; savM = Minutes; savS = Secondes; Heure = 2; Minutes = 0; Secondes = 0; Azimut = 30;
  Pos_X_sur_ecran = map(Colonne,2,109,43,315); // Pos_X_sur_ecran contient Azimut correspondant à Colonne.
  Effacer_zone(4, 20, 3); display.setCursor(4,20); display.print(Pos_X_sur_ecran); display.update(); 
  //================= Attendre d'être à l'Azimut du curseur ================== 
  while (Azimut < Pos_X_sur_ecran) 
        {posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude,
                   &Hauteur, &Azimut); Faire_evoluer_le_temps();}
  //================ Afficher l'heure correspondante =================
  Effacer_zone(75, 4, 8); AFF_HEURE(); display.update();
  //============================================================================
  Heure = savH; Minutes = savM; Secondes = savS;}

void Traite_BP_Central() {
  if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
     //============================== Traiter le mode GRAPHIQUE ==============================
     digitalWrite(LED_Arduino, HIGH); // Allume la LED.
     while (analogRead(BP_central) < 512); // Attendre le relâché.
     digitalWrite(LED_Arduino, LOW);
     Mode_trois_courbes = false; Aff_graduations = false; Affiche_ecran_Graphique();
     OCTET = 1; // OCTET ici représente le coef de déplacement latéral.
     Colonne = 2; // Positionne le curseur pour le bouton rotatif.
     Affiche_le_curseur(); Affiche_coef_deplacement(); Clic_long = false; 
     while (!Clic_long) {
       //===================== Tester le clavier à deux B.P. ==================
       Tester_le_BP_A0();
       if (Clic_sur_le_BP && !Clic_long) {
           Clic_sur_le_BP= false;
           if (Augmenter_le_choix) {Mode_trois_courbes = !Mode_trois_courbes; Affiche_ecran_Graphique();}
           else {Aff_graduations = !Aff_graduations; Affiche_ecran_Graphique();} }
       //======================= BP central en sous fonction ==================
       if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
           digitalWrite(LED_Arduino, HIGH); // Allume la LED.
           while (analogRead(BP_central) < 512); // Attendre le relâché.
           digitalWrite(LED_Arduino, LOW);
           OCTET = OCTET << 1; if (OCTET==16) OCTET=1;
           Affiche_coef_deplacement();}           
       //========================= Traite la rotation =========================
       INT_active = true; // Armer un nouveau comptage.
       if (Rotation) {
          Rotation = false;       
          if (Mode_trois_courbes) BIP();
          else {Efface_le_curseur();
                if (Sens_positif) {Colonne = Colonne + OCTET; if (Colonne > 109) Colonne = 2;}
                else {if (Colonne < OCTET+2) Colonne = 109; else Colonne = Colonne - OCTET;}
                Affiche_le_curseur();
                Determine_parametres_pour_le_curseur();} } }
  //============================= Sortie du mode GRAPHIQUE ===============================
  Clic_sur_le_BP = false; Effectuer_les_calculs(); Affiche_Position_Soleil(); Mode_Ecran = 1;} }

void Modifie_element() {INT_active = true; // Armer un nouveau comptage.
  int MAXIMUM, VALEUR;
  INT_active = true; // Armer un nouveau comptage.
  if (Rotation) { Rotation = false;
     //-------------------------- Positionner le maximum possible --------------------------
     if (Edite_HEURE)
          switch (Mode_Edition)
            {case 0 : {VALEUR = Heure; MAXIMUM = 23; break;}
             case 1 : {VALEUR = Minutes; MAXIMUM = 59; break;}
             case 2 : {VALEUR = Secondes; MAXIMUM = 59;}}
     else switch (Mode_Edition)
            {case 0 : {VALEUR = Jour; MAXIMUM = 31; break;}
             case 1 : {VALEUR = Mois; MAXIMUM = 12; break;}
             default : VALEUR = Annee; MAXIMUM = 3000; break;}  
     //----------------------------- Calcule l'élément pointé ------------------------------
        if (Sens_positif) {VALEUR++; if (VALEUR > MAXIMUM) VALEUR = 0;
            if (VALEUR == 0 && !Edite_HEURE) VALEUR = 1;}
        else {if (VALEUR > 0) VALEUR--; else VALEUR = MAXIMUM;}
        if (VALEUR==0 && !Edite_HEURE) VALEUR = MAXIMUM;
      //------------------------- Mise à jour de l'élément pointé --------------------------
     if (Edite_HEURE)
          switch (Mode_Edition)
            {case 0 : {Heure = VALEUR; break;}
             case 1 : {Minutes = VALEUR; break;}
             default : {Secondes = VALEUR;}}
     else switch (Mode_Edition)
            {case 0 : {Jour = VALEUR; break;}
             case 1 : {Mois = VALEUR; break;}
             default : Annee = VALEUR; }
     SAVM = Mois; SAVJ = Jour;   
     Effacer_zone(7, 4, 19); Mise_a_jour_Date_et_Heure(); display.update();}}
     
void Accent_aigu(byte X, byte Y) {display.drawLine(X, Y, X+1, Y-1, WHITE);}

void Changer_Date_Observation() {
  byte Date_speciale = 0;
  Mode_Edition = 0; Edite_HEURE = false; // Donc on édite la date.
  Affiche_date_et_heure(); Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(174, 16);
  display.setCursor(5,36); Aff_TEXTE_EEPROM(192, 19); Accent_aigu(109,36); Positionne_curseur_DATE();
  Clic_long = false;
  while (!Clic_long) {
    Tester_le_BP_A0();
    if (Clic_sur_le_BP && !Clic_long) {
         Clic_sur_le_BP= false;
         if (Augmenter_le_choix) {
             Clic_sur_le_BP= false;  Mode_Edition++; if (Mode_Edition == 3) Mode_Edition = 0;
             Positionne_curseur_DATE();}
         else {savJ = Jour; savMois = Mois; Effacer_zone(16, 48, 18); // Efface le pointeur.
               display.setCursor(16,48); // Pour afficher la nature de la date sélectionnée. 
               switch (Date_speciale)
                  {case 0 : {Mois = 3; Jour = 21; Aff_TEXTE_EEPROM(212, 17); break;} // Equinoxe Printmps
                   case 1 : {Mois = 6; Jour = 21; Aff_TEXTE_EEPROM(230, 13);
                             Accent_aigu(84,49); break;} // Solstice Eté.
                   case 2 : {Mois = 9; Jour = 23; Aff_TEXTE_EEPROM(244, 16); break;} // Equinoxe Automne.
                   case 3 : {Mois = 12; Jour = 21; Aff_TEXTE_EEPROM(261, 15); break;} // Solstice Hiver.
                   case 4 : {Accent_aigu(90,49); Aff_TEXTE_EEPROM(278, 16);} } // > Jour courant <             
             Effacer_zone(7, 4, 19); Mise_a_jour_Date_et_Heure(); display.update();
             Date_speciale++; if (Date_speciale == 5) Date_speciale = 0;} }        
    Modifie_element();} //Si Bouton rotation modifie la valeur et affiche les changements.
  Clic_sur_le_BP= false; Effectuer_les_calculs(); Affiche_Position_Soleil(); Mode_Ecran = 1;}
  
void Changer_Heure_Observation() {
  Mode_Edition = 0; Edite_HEURE = true; // Donc on édite la date.
  Affiche_date_et_heure(); Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(297, 16);
  display.setCursor(5,36); Aff_TEXTE_EEPROM(312, 17); Positionne_curseur_DATE();
  Clic_long = false;
  while (!Clic_long) {
    Tester_le_BP_A0();
    if (Clic_sur_le_BP && !Clic_long) {
         Clic_sur_le_BP= false;  Mode_Edition++; if (Mode_Edition == 3) Mode_Edition = 0;
         Positionne_curseur_DATE();}
    Modifie_element(); } //Si Bouton rotation modifie la valeur et affiche les changements.
  Clic_sur_le_BP= false; Effectuer_les_calculs(); Affiche_Position_Soleil(); Mode_Ecran = 1;}

void Positionne_curseur_DATE() {
  Effacer_zone(14, 48, 17); // Efface le pointeur ou texte. 
  display.setCursor(16,48); if (Mode_Edition ==1) display.setCursor(54,48);
  if (Mode_Edition ==2) display.setCursor(100,48); display.print("^"); display.update();}
  
void Positionne_curseur_ALTITUDE() {
  Effacer_zone(12, 48, 19); // Efface le pointeur. 
  display.setCursor(16,48);
  switch (Mode_Edition)
         {case 0 : {display.setCursor(14,48); break;}
          case 1 : {display.setCursor(53,48); break;}
          case 2 : {display.setCursor(92,48); break;}
          default : display.setCursor(119,48);}
  display.print("^"); display.update();}
  
void Rafraichi_Altitude() {
  Effacer_zone(50, 4, 6); display.setCursor(50,4); display.print(Altitude);
  display.print(" m"); display.update();} 
    
void Rafraichir_type() {// procedure locale
    Effacer_zone(65, 20, 9); display.setCursor(65,20);
    if (Edit_Longitude) Aff_TEXTE_EEPROM(330, 9); else Aff_TEXTE_EEPROM(341, 8); display.update();}
    
void AFF_E_W() {if (Longitude_EST) display.print(" E"); else display.print(" W");}

void Rafraichir_Position() {
  Effacer_zone(31, 4, 11); display.setCursor(31,4); if (Edit_Longitude) Affiche_1_espace(); 
  if (Edit_Longitude) {display.print(Longitude,5); AFF_E_W();}
  else {display.print(Latitude,5); display.print(" N");} }

void Positionne_curseur_Coordonnees() {
  Effacer_zone(37, 48, 7); // Efface le pointeur. 
  display.setCursor(37,48); display.setCursor(37 + (Mode_Edition * 6),48);
  display.print("^"); display.update();}

void Changer_le_lieu_Observation() {
  //=================================== Changer l'ALTITUDE ======================================
  Mode_Edition = 0; Edite_HEURE = true; // Donc on édite la date.
  display.clear(); Rafraichi_Altitude();
  Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(353, 18);
  display.setCursor(5,36); Aff_TEXTE_EEPROM(371, 20); Positionne_curseur_ALTITUDE();
  Clic_long = false;
  while (!Clic_long) {
    Tester_le_BP_A0();
    if (Clic_sur_le_BP && !Clic_long) {
         Clic_sur_le_BP= false;  Mode_Edition++; if (Mode_Edition == 4) Mode_Edition = 0;
         Positionne_curseur_ALTITUDE();}
    //-------------------------- Traite la rotation --------------------------
    INT_active = true; // Armer un nouveau comptage.
    if (Rotation) { Rotation = false;
       if (Sens_positif) {        
       switch (Mode_Edition) // Maximum 9999m limité par l'affichage de la page des données d'observation.
              {case 0 : {Altitude = Altitude + 1000; break;}
               case 1 : {Altitude = Altitude + 100; break;}
               case 2 : {Altitude = Altitude + 10; break;}
              default : Altitude++;}; if (Altitude > 9999){ Altitude = 9999; BIP();}}
     else{switch (Mode_Edition) // Minimum zéro limité par l'affichage de la page des données.
              {case 0 : {Altitude = Altitude - 1000; break;}
               case 1 : {Altitude = Altitude - 100; break;}
               case 2 : {Altitude = Altitude - 10; break;}
              default : Altitude--;}; if (Altitude < 0){ Altitude = 0; BIP();}}       
       Rafraichi_Altitude();}
     //-----------------------------------------------------------------------
  Clic_sur_le_BP= false;}
  //============================== Changer les COORDONNEES GPS ==================================
  Edit_Longitude = true; Mode_Edition = 0; 
  display.clear(); Rafraichir_Position();
  Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(352, 9); Rafraichir_type();
  display.setCursor(37,36); Aff_TEXTE_EEPROM(393, 7); Positionne_curseur_Coordonnees();
  Clic_long = false;
  while (!Clic_long) {
    Tester_le_BP_A0();
    if (Clic_sur_le_BP && Clic_long && !Augmenter_le_choix) {
        Clic_long = false; Clic_sur_le_BP = false;
        if (Edit_Longitude) { Longitude_EST = !Longitude_EST; Rafraichir_Position(); display.update();} else BIP();} 
    if (Clic_sur_le_BP && !Clic_long) {
         Clic_sur_le_BP= false;
         if (Augmenter_le_choix) 
            {Mode_Edition++;
             if (Mode_Edition == 7) Mode_Edition = 0; if (Mode_Edition == 1) Mode_Edition = 2;
         Positionne_curseur_Coordonnees();}
         else {Edit_Longitude = !Edit_Longitude; Rafraichir_Position(); Rafraichir_type();
               Mode_Edition = 0; Positionne_curseur_Coordonnees();}}
    //-------------------------- Traite la rotation --------------------------
    INT_active = true; // Armer un nouveau comptage.
    if (Rotation) { Rotation = false;
       if (Edit_Longitude) VALEUR = Longitude; else VALEUR = Latitude;
       if (Sens_positif) {        
       switch (Mode_Edition)
              {case 0 : {VALEUR++; break;}
               case 2 : {VALEUR = VALEUR + 0.1; break;}
               case 3 : {VALEUR = VALEUR + 0.01; break;}
               case 4 : {VALEUR = VALEUR + 0.001; break;}
               case 5 : {VALEUR = VALEUR + 0.0001; break;}
               case 6 :  VALEUR = VALEUR + 0.00001;} // 6 obligé car 1 ignoré. (Point décimal)
               Traite_limite_SUP();}
       else {switch (Mode_Edition)
              {case 0 : {VALEUR--; break;}
               case 2 : {VALEUR = VALEUR - 0.1; break;}
               case 3 : {VALEUR = VALEUR - 0.01; break;}
               case 4 : {VALEUR = VALEUR - 0.001; break;}
               case 5 : {VALEUR = VALEUR - 0.0001; break;}
               case 6 :  VALEUR = VALEUR - 0.00001;} // 6 obligé car 1 ignoré. (Point décimal)
               Traite_limite_INF();}
     if (Edit_Longitude) Longitude = VALEUR; else Latitude = VALEUR;  
     Rafraichir_Position(); display.update();}
     //-----------------------------------------------------------------------  
  Clic_sur_le_BP= false;}  
  //=============================================================================================
  Effectuer_les_calculs(); Affiche_Position_Soleil(); Mode_Ecran = 1;}
  
void Traite_limite_SUP() {
  if (Edit_Longitude && VALEUR > 9) {VALEUR = 9; BIP();}
  if (!Edit_Longitude && VALEUR > 52) {VALEUR = 52; BIP();} }

void Traite_limite_INF() {
  if (Edit_Longitude && VALEUR < 0) {VALEUR = 0; BIP();}
  if (!Edit_Longitude && VALEUR < 40) {VALEUR = 40; BIP();} }

void Augmente_mode_crepuscule() {
  Mode_Crepuscule = Mode_Crepuscule + 6; if (Mode_Crepuscule == 24) Mode_Crepuscule = 0;}
  
void Changer_mode_crepusculaire() {
  while (!Clic_long) {
        Affiche_mode_crepusculaire();
        if (Rotation) {Rotation = false;
            if (Sens_positif) Augmente_mode_crepuscule();
            else {if (Mode_Crepuscule > 0) Mode_Crepuscule = Mode_Crepuscule - 6;
                  else Mode_Crepuscule = 18; }}
        Tester_le_BP_A0();
        if (Clic_sur_le_BP && !Clic_long) {BIP(); Clic_sur_le_BP = false;} }
  Clic_sur_le_BP = false;
  Effectuer_les_calculs(); Affiche_Lever_Coucher(); Mode_Ecran = 2;}

void Affiche_mode_crepusculaire() {
  display.clear(); Affiche_cadre_du_haut(12); Aff_TEXTE_EEPROM(401, 18);
  CADRE(); display.setCursor(7,20); Aff_TEXTE_EEPROM(420, 19); Accent_aigu(74,21);
  display.drawLine(1, 30, 126, 30, WHITE); 
  display.setCursor(5,42); 
  if (Mode_Crepuscule==0) display.setCursor(5,38); Aff_TEXTE_EEPROM(440, 7);
  switch (Mode_Crepuscule) {
        case  0 : {Aff_TEXTE_EEPROM(448, 9); display.setCursor(5,48);
                   Aff_TEXTE_EEPROM(458, 20); Accent_aigu(72,49); break;} 
        case  6 : {Aff_TEXTE_EEPROM(479, 6); break;}
        case 12 : {Aff_TEXTE_EEPROM(486, 9); break;}
        case 18 : {Aff_TEXTE_EEPROM(496, 13); break;}}
  display.update(); }

void Efface_ecran() {
  display.clear(); display.setTextSize(2); display.setCursor(15,20); Aff_TEXTE_EEPROM(637, 8);
  display.update(); delay(1000); display.clear(); display.update();  display.setTextSize(1);}

//-------------------- Routines de servitude pour traiter le capteur incrémental -------------------
void Delay_antirebonds() {for(int I=0; I<3; I++);} // Valeur de 3 expérimentale.

void Front_sur_A(){ // Interruption quand A change d'état avant B.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
  if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
    A_set = !A_set;
    if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }

void Front_sur_B(){ // Interruption quand B change d'état avant A.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
  if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
    B_set = !B_set;
    // Décrémenter le compteur si B avant A.
    if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
//--------------------------------------------------------------------------------------------------
  
void Affiche_la_Position() {
  Affiche_date_et_heure(); Prepare_cadre_3_donnees(); display.drawLine(1, 30, 62, 30, BLACK);
  Aff_TEXTE_EEPROM(510, 14); display.print(Altitude); display.print("m");
  display.drawLine(63, 17, 63, 30, WHITE); display.setCursor(10,36); Aff_TEXTE_EEPROM(341, 9);
  display.print(Latitude,5); display.print(" N");
  display.setCursor(10,48); Aff_TEXTE_EEPROM(330, 10); display.print(Longitude,5);AFF_E_W(); 
  display.update();} 

void Affiche_Lever_Coucher() {
  Affiche_date_et_heure();
  display.drawLine(71, 1, 126, 14, WHITE); display.drawLine(71, 14, 126, 1, WHITE); // Barre l'heure.
  Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(525, 10); display.print(Culmine);
  display.setCursor(10,36); Aff_TEXTE_EEPROM(578, 10); display.print(Lever); display.setCursor(10,48);
  Aff_TEXTE_EEPROM(589, 10); display.print(Coucher); display.update();} 

void Affiche_Position_Soleil() {
  Affiche_date_et_heure(); Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(536, 19);
  display.setCursor(15,36); Aff_TEXTE_EEPROM(556, 10); display.print(Hauteur, 3);
  display.setCursor(15,48); Aff_TEXTE_EEPROM(567, 10); display.print(Azimut, 3); display.update();}
 
void Prepare_cadre_3_donnees() {
  CADRE(); display.drawLine(1, 30, 126, 30, WHITE); display.setCursor(9,20); } 

void AFF_HEURE() {display.setCursor(75,4); Complete_a_10(Heure); display.print(Heure); Affiche_separation();
  Complete_a_10(Minutes); display.print(Minutes); Affiche_separation(); Complete_a_10(Secondes);
  display.print(Secondes);}
  
void Affiche_la_date() {
  Complete_a_10(Jour); display.print(Jour); Affiche_1_espace(); Complete_a_10(Mois); display.print(Mois);
  Affiche_1_espace(); display.print(Annee);} 
  
void Mise_a_jour_Date_et_Heure()
  {Affiche_cadre_du_haut(7); Complete_a_10(Jour); display.print(Jour); Affiche_1_espace();
  Complete_a_10(Mois); display.print(Mois); Affiche_1_espace(); display.print(Annee); AFF_HEURE();
  display.drawLine(70, 1, 70, 14, WHITE);}

void Affiche_date_et_heure() {display.clear(); Mise_a_jour_Date_et_Heure();}

void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {
  for (byte I=0; I<Longueur; I++) {display.print(char(EEPROM.read(PTR))); PTR++;}; }

void Complete_a_10(int Valeur) {if(Valeur < 10) display.print("0");}
void Affiche_separation() {display.print(":");}
void Affiche_1_espace() {display.print(" ");}
void Affiche_N_espaces(byte N) {for (byte I = N; I > 0; I--) Affiche_1_espace();}

void Effacer_zone(byte X, byte Y, byte NB_caracteres) {
  byte I = 10; 
  for (byte J=0; J < I; J++)
       display.drawLine(X, Y + J, X + (NB_caracteres * 6), Y + J, BLACK);}
  
void Affiche_cadre_du_haut(byte X) {
  CADRE_UP();  display.setTextSize(1); // Taille la plus petite.
  display.setCursor(X,4);}

void CADRE_UP() {display.drawRect(0, 0, 128, 16, WHITE);}
void CADRE() {display.drawRect(0, 16, 128, 48, WHITE);}

void BIP() {
  pinMode(BUZZER, OUTPUT); tone(BUZZER,3000); delay(100); noTone(BUZZER); pinMode(BUZZER, INPUT);}
  
void ATTENDRE_un_BP() { Clic_sur_le_BP = false;
  while (!Clic_sur_le_BP) {
    digitalWrite(LED_Arduino,!digitalRead(LED_Arduino)); // Inverser la LED clavier.
    delay(50); Tester_le_BP_A0();}
  digitalWrite(LED_Arduino,LOW); Clic_sur_le_BP = false; }
  
// >>>>>>>>>>>>>>> Le B.P. est câblé en "Analogique" <<<<<<<<<<<<<<<<<<
void Tester_le_BP_A0() { // La lecture sur A0 donne 0 ou 930 maximum.
  int Numerisation_BP; // Valeur retournée par la CAN de A0.
  boolean Sauve_etat_LED_jaune;
  byte COMPTEUR; unsigned long Horloge_actuelle;
  Sauve_etat_LED_jaune = digitalRead(LED_Jaune); Numerisation_BP = analogRead(0);
  if ((Numerisation_BP > 710) || (Numerisation_BP < 240)) {
    Clic_sur_le_BP = true; digitalWrite(LED_Arduino, HIGH);
    Augmenter_le_choix = false;
    Horloge_actuelle = millis(); Clic_long = false;
    if (Numerisation_BP > 710) { Augmenter_le_choix = true; }
    delay(100); // Attendre_la_stabilisation du travail.
    // Attendre_la_stabilisation du repos :
    COMPTEUR = 0;
    while (COMPTEUR < NB_tests_antirebonds) {
      if ((millis() - Horloge_actuelle) > 700) {Clic_long = true; digitalWrite(LED_Jaune, HIGH);}
      COMPTEUR++; Numerisation_BP = analogRead(0);
      if ((Numerisation_BP > 710) || (Numerisation_BP < 240)) COMPTEUR = 0; }
  digitalWrite(LED_Arduino, LOW);} // Éteindre la LED clavier.
  digitalWrite(LED_Jaune, Sauve_etat_LED_jaune); } // Restitue l'état de la LED jaune.
  
