
/*  NANO Éphéméride Solaire avec affichage sur un module OLED SSD1306.
           Par Michel DROUI. (Pseudo Internet : Nulentout.)
           
    >>>>>> ATTENTION : Il ne reste que 6 octetes de disponibles en EEPROM.
    >>>>>> ATTENTION : Il ne reste que 128 octetes de disponibles pour le programme.
    >>>>>> Pour pouvoir afficher la RAM disponible pour la pile il peut s'avérer impératif
           valider les deux lignes du GOTO qui font sauter l'affichage du LOGO.
    >>>>>> Affiche une zone de RAM dynamique confortable, soit : 519 Octets.
           
    La latitude est comprise entre 40° et 55°.
    Les longitudes possibles sont imitées entre -9° et +9°.
    L'ALTITUDE est limitée entre zéro et 9999m.    
    Détermine le changement d'heure avec une imprecision annuelle de + ou - 3 Jours maximum.
           
    Le B.P. central est affecté au MODE GRAPHIQUE.
    Le bouton rotatif hors mode graphique fait changer de mode écran :
          0 : Mode veille > Écran noir.
          1 : Affichage de la position du Soleil.
          2 : Affichage du Lever et du Coucher.
          3 : Affichage de l'Équation du temps.
          4 : Affichage d'informations complémentaires : Jour Julien, durée de la nuit ...
          5 : Affichage du mode crépusculaire.
          6 : Affichage de la situation géographique.
    FC- court : Changer le Mode crépusculaire.
    FC+ court : Changer la DATE d'Observation.
                FC+ court change de donnée.
                FC- court recircule des dates particulières.         
    FC- long : Changer le LIEU de l'observatoire. 
    FC+ long : Changer l'HEURE d'Observation.        

DIFFERENTS BRANCHEMENTS :
 D0  : Libre pour F.E.                         A0 : Lecture du clavier à deux B.P. FC- et FC +.
 D1  : Libre pour F.E.                         A1 : Lecture du bouton central du capteur rotatif.
 D2  : Codeur Incrémental A.                   A2 : Mesure de la tension sur l'accumulateur d'alimentation.
 D3  : Codeur Incrémental B.                   A3 : Mesure de la tension sur la PILE de sauvegarde.    
 D4  : RST HORLOGE.                            A4 : Libre pour F.E.
 D5  : I/O HORLOGE.                            A5 : Libre pour F.E.
 D6  : SCLK HORLOGE.  
 D7  : LED rouge d'avertissements divers.      A6 : Libre pour F.E.
 D8  : Pilotage du BUZZER.                     A7 : Libre pour F.E.
 D9  : LED jaune CLAVIER à deux B.P.
 D10 : Libre pour F.E.                         NOTE : Certaines broches ont été affectées de
 D11 : Broche_SDA Pilotage OLED.                      façon à simplifier l'étude du circuit imprimé.
 D12 : Broche_SCL Pilotage OLED.
 D13 : Libre pour F.E. 

   CE QUI A ÉTÉ AJOUTÉ :
   =====================
- Intégration de l'Horloge électronique DS1302.
- Mise à jour de l'Horloge.
- Mode affichage position en temps réel.
  >>> Culmine, Lever, et Coucher sont indiqués :
      Pour le temps réel quand on progresse de "Position du Soleil" vers le mode d'affichage des crépuscules.
      Pour la date et heure simulées quand on "rétrograde de "Position simulée" > VEILLE vers le mode.
      Dans les deux cas DATE et HEURE sont indiquées ce qui évite toute ambiguîté.
- Mode affichage de l'équation du temps.
  >>> L'index de position, le n° du jour et la valeur de l'E.T. sont indiqués :
      Pour le temps réel quand on progresse de "Position du Soleil" vers le mode d'affichage des crépuscules.
      Pour la date simulées quand on "rétrograde de "Position simulée" > VEILLE vers le mode.
      Dans ce mode d'affichage :
           FC+ et FC- courts font balayer l'année.
           FC+ long change le coefficient de rapidité de balayage.
           FC- long alterne entre Temps réel et Date simulée.
- Le mode d'affichage de données complémentaires. (Jour Julien, mesure des alimentations etc.
  Un B.P. court dans ce mode alerterne entre deux pages d'informations.
*/

// Avant intégration de l'HORLOGE DS1302 : 23688 (77%) 319 (15¨%)
// Avec Mise à l'haure ajoutée :           26276 (85%) 330 (16 %)
// Avec Affichage en temps réel ajoute   : 26400 (85%) 332 (16 %)
// Avec Affichage de l'équation du temps : 28016 (91%) 390 (19 %)
// Avec Affichages complémentaires       : 30604 (99%) 406 (19 %)
// Après optimisation du code            : 30262 (98%) 406 (19 %)
// Après optimisation des textes EEPROM  : 30144 (98%) 386 (18 %)
// Après gestion des constellations      : 30542 (99%) 388 (18 %)
// Après surveillance de l'Accumulateur  : 30592 (99%) 432 (21 %)

#include <Adafruit_ssd1306syp.h> // Impérativement avant #include <EEPROM.h>.

// Déclarations pour le LOGO.
#include <EEPROM.h>
#define XL 8  // Position du LOGO.
#define YL 34 // Position du LOGO.

#include <stdio.h>  // Impératif pour l'HORLOGE DS1302.
#include <DS1302.h> // Bibliothèque pour l'HORLOGE DS1302.
// Dédinir les E/S sur Arduino.
uint8_t LigneRST  = 4;
uint8_t LigneIO   = 5;
uint8_t LigneSCLK = 6;
// Creer "l'objet" DS1302.
DS1302 rtc(LigneRST, LigneIO, LigneSCLK);

// Déclarations pour l'afficheur OLED :
#define broche_SDA 11 // D11 utilisée en sortie.
#define broche_SCL 12 // D12 utilisée en sortie.
Adafruit_ssd1306syp display(broche_SDA, broche_SCL);

// Déclarations pour le capteur codeur rotatif KY-40 :
#define Codeur_B 3    // D2 : INT 0 sur broche 4 du microcontrôleur. (DT)
#define Codeur_A 2   // D3 : INT 1 sur broche 5 du microcontrôleur. (CLK)
#define BP_central 15 // A1 pour lire le bouton central.

// Constantes du programme.
#define Version_PGM "Version 31-05-17" // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Coef_CAN_pour_la_pile 0.96 // Correction du convertisseur A.N. pour la mesure de la pile de sauvegarde.
#define Coef_CAN_pour_accumulateur 1 // Correction du convertisseur A.N. mesure accumulateur d'alimenlation.
#define Tension_Pile_minimale 2.8 // Tension à partir de laquelle il y aura alerte.
#define Tension_Accu_minimale 6.5 // Tension à partir de laquelle il y aura alerte.
double Latitude = 44.4491;  // (Arcambal)
double Longitude = 1.43663;  // (Arcambal)
int Altitude = 142; // Altitude d'observation. (142 pour Arcambal)

#define Max_mode_ecran 8 // Nombre de modes écran - 1.
#define LED_Jaune 7  // LED d'avertissements divers.
#define BUZZER    8  // Sortie dédiée au BUZZER.
#define LED_Clavier 9
#define NB_tests_antirebonds 50 // Nombre de tests de stabilisation B.P.

// Variables du programme.
boolean Clic_sur_le_BP, Augmenter_le_choix, Clic_long, BP_central_a_traiter; 
volatile boolean Rotation;  // Indique une activation du codeur rotatif.
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean Sens_positif;   // Sens de rotation.
boolean Edite_HEURE;    // On édite soit l'HEURE, soit la DATE.
boolean Aff_graduations; // Affiche les graduations en mode graphique.
boolean Mode_trois_courbes, Edit_Longitude, Affiche_date;
boolean Longitude_EST;
boolean Temps_reel = true; // Si VRAI rafraichi en permanence DATE et HEURE;
boolean E_T_en_temps_reel;
boolean AFF_Info_Complementaire;
boolean Non_test_accumulateur;
byte OCTET, Ligne; // Pour le LOGO.
byte X; // Position sur X pour les titres et pour le LOGO.
byte Y; // Position sur Y pour tracer les pixels.
byte PTR; // Pour le LOGO et le pointeur en RAM d'échantillons.
byte Mode_Ecran = 2; // Sur RESET afficher en temps réel la position solaire.
byte Coef_deplacement, Mode_Edition;
byte Mode_Crepuscule = 0;
	//   -  0 : Crépuscule sur l'horizon avec correction de la réfraction athmosphérique soit 0.833°
	//   -  6 : Crépuscule civil
	//   -  12 : Crépuscule nautique
	//   -  18 : Crépuscule astronomique
byte Mois, Jour, Heure, Minutes, Secondes, SAVM, SAVJ, SAVH, SAVmin, SAVs ;
int  SAVA, Annee; 
byte savH, savM, savS, savJ, savMois;
int Pos_X_sur_ecran;
int Num_Jour_dans_annee;
int Colonne; // Int car utilisé pour "Jour" allant jusqu'à 365;
long Valeurs_lues;
float VALEUR;
float Tension_lue;

//***************************** Déclarations spécifiques aux calcules astronomiques ***************************
#define Nombre_de_degres_par_Radian 57.295779505601 // Nombre de degrés par Radian.
boolean Heure_Hiver;
boolean Bissextile;
float Equation_du_Temps; // Float car peut être négatif et décimal.
int nbJ_par_mois[] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
float declinaison; // Déclinaison en radians.
float Hauteur;
float Azimut;
String Lever, Culmine, Coucher;
float Lever_en_min, Coucher_en_min;
//*************************************************************************************************************

void setup() {
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  pinMode(LED_Clavier, OUTPUT);  digitalWrite(LED_Clavier, LOW);
  pinMode(LED_Jaune, OUTPUT); digitalWrite(LED_Jaune, LOW); pinMode(BUZZER, INPUT);
  digitalWrite(BP_central, HIGH); // Valide R PUL-UP sur A1.
  display.initialize(); delay(300); display.clear();
  display.setTextSize(1); // Taille la plus petite.
  display.setCursor(18,4);
  display.print(Version_PGM); 
  CADRE(); display.setTextSize(2); // Taille double.
  display.setCursor(47,21); Aff_TEXTE_EEPROM(120, 6); 
  //================================= Tracer le LOGO. ===============================
  //goto SUITE; //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  PTR = 0; // Pointer le premier octer dans l'EEPROM.
  for(Ligne = 0; Ligne < 24; Ligne++)
     {for(Colonne = 0; Colonne < 5; Colonne++) 
         {OCTET = EEPROM.read(PTR);
          for(byte I=0; I < 8; I++)
              {display.drawPixel((Colonne*8)+XL+I, Ligne+YL, (OCTET & B10000000));
               OCTET=OCTET<<1;}; PTR++; } };
  display.drawLine(XL+40, YL+18, XL+40, YL+22,WHITE); display.drawLine(XL+41, YL+19, XL+41, YL+21,WHITE);
  //SUITE: //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ 
  //=================================================================================
  display.setTextSize(1); display.setCursor(70,45); Aff_TEXTE_EEPROM(144, 5); 
  display.update();
  ATTENDRE_un_BP(); // Attend un BP pour sortir. @@@@@@@@@@@@@@@@@ Valider ajoute 68 Octets @@@@@@@@@@@@@@@@@@@@@
  // --------------- mettre à jour DATE et HEURE et mémoriser ces données ---------------
  Transfere_Donnees_DS1302(); // Mise à jour DATE/HEURE depuis l'horloge interne.
  SAVA = 2017; SAVM = 1; SAVJ = 1; SAVH = 12; SAVmin = 0; SAVs = 0;
  Longitude_EST = true; Effectuer_les_calculs(); // Avoir des valeurs cohérentes au démarrage.
//@@@@@@@@@@@ Ci-dessous code ajouté pour afficher la place disponible. @@@@@@@@@@@
//  display.clear(); display.setCursor(30,30); display.print(SRAM_LIBRE());
//  display.update(); ATTENDRE_un_BP(); // vérifier la non "collision de PILE".
//@@@@@@@@@@@@**********************************************************@@@@@@@@@@@
  Non_test_accumulateur = false; // Option sur RESET : Test de l'accumulateur. @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  BP_central_a_traiter = false; Teste_la_PILE(); Teste_Accumulateur();}

//============ Ci-dessous fonction pour afficher la place disponible ============
// Inutile de passer cette fonction en remarque car si la séquence de mesure de
// la RAM dynamique n'est pas validée ce code n'est pas compilé par l'IDE.
  int SRAM_LIBRE() { // Fonction qui retourne la taille de SRAM disponible.
  extern int __heap_start, *__brkval; // Déclaration des deux pointeurs dédiés.
  byte BIDON; // Dernière variable allouée, donc occupe le "haut" du TAS.
  if (__brkval == 0) {return (int) &BIDON -(int) &__heap_start;}
  else {return (int) &BIDON -(int) __brkval;} }
//===============================================================================  

void loop() {
  if ((Temps_reel) && (Mode_Ecran==2))
     {Transfere_Donnees_DS1302(); // Ne modifie pas Date/heure simulées.
      Effectuer_les_calculs(); Affiche_Position_Soleil();} 
  //======================== Traite le B.P. central ======================
  Teste_activation_BP_central(); if (BP_central_a_traiter) {BP_central_a_traiter = false; Traite_BP_Central();}
  //========================= Traite la rotation =========================
  INT_active = true; // Armer un nouveau comptage.
  if (Rotation) { Rotation = false;
  AFF_Info_Complementaire = true; // réarme la bascule d'affichage des Informations complémentaires.
     if (Sens_positif) {Mode_Ecran++; if (Mode_Ecran > Max_mode_ecran) Mode_Ecran = 0;}
     else {if (Mode_Ecran==0) Mode_Ecran = Max_mode_ecran;  else Mode_Ecran--;}
     Effectuer_les_calculs(); Teste_la_PILE(); // Teste la pile à chaque changement de fonction.
     Teste_Accumulateur(); // Teste l'Accumulateur 9V à chaque changement de fonction.
     switch (Mode_Ecran) {
        case 0 : {Affiche_HORLOGE_temps_reel(); break;} // Horloge temps réel.
        case 1 : {Temps_reel=false; Restituer_Simulation(); Effectuer_les_calculs(); // Simulation.
                  Affiche_Position_Soleil(); Efface_Zone_et_Set_Curseur(60, 20, 10,60,20);
                  Aff_TEXTE_EEPROM(227, 5); display.print(char(130)); display.print("e."); display.update(); break;}
        case 2 : {Temps_reel=true; Affiche_Position_Soleil(); break;} // Affichage en temps réel.
        case 3 : {Affiche_Lever_Coucher(); break;} // Affichage du Lever et du Coucher.
        case 4 : {Coef_deplacement = 1; Trace_Equation_du_temps(); break;} // Affichage l'équation du temps.
        case 5 : {Affichages_complementaires(); break;} // Durée du jour, heure Été/Hiver ...        
        case 6 : {Changer_mode_crepusculaire(); break;} // Affichage du mode crépusculaire.
        case 7 : {Affiche_la_Position(); break;} // Affichage de la situation géographique.
        case 8 : Efface_ecran(); }} // Mode veille > Écran noir.
  //=============== Traiter la requête de l'un des deux B.P. =============
  Tester_le_BP_A0();
  if (Clic_sur_le_BP) {Clic_sur_le_BP = false; // Action sur le B.P. prise en compte.
      if (Mode_Ecran == 4)
      //==================== Écran sur affichage Équation du temps ====================
         //================ Ci-dessous traitement du clic long =================
         if (Clic_long) {
             if (Augmenter_le_choix) {
                 switch (Coef_deplacement) {
                     case 30 : {Coef_deplacement = 1; break;}
                     case 15 : {Coef_deplacement = 30; break;}
                     case 7 : {Coef_deplacement = 15; break;}
                     case 4 : {Coef_deplacement = 7; break;}
                     case 2 : {Coef_deplacement = 4; break;}
                     case 1 : Coef_deplacement = 2;}
                 Afficher_coef_deplacement();}
             else {E_T_en_temps_reel = !E_T_en_temps_reel; Affiche_date = true;
                   if (E_T_en_temps_reel) Transfere_Donnees_DS1302(); else Restituer_Simulation();
                   Num_Jour_dans_annee = Jour_Num(Mois, Jour);
                   Afficher_Jour_valeur_ET_Curseur_Bissextile();}}
         //================ Ci-dessous traitement du clic court ================
         else {if (Augmenter_le_choix) 
                  {Num_Jour_dans_annee = Num_Jour_dans_annee + Coef_deplacement;
                   if (Num_Jour_dans_annee > 365) Num_Jour_dans_annee = 1;}
               else {if (Num_Jour_dans_annee >= Coef_deplacement)
                         Num_Jour_dans_annee = Num_Jour_dans_annee - Coef_deplacement;
                     else Num_Jour_dans_annee = 365;}
               Affiche_date = false; Afficher_Jour_valeur_ET_Curseur_Bissextile();}                     
      //=============== Écran sur affichage Informations complémentaires ==============
      if (Mode_Ecran == 5)
         {AFF_Info_Complementaire = !AFF_Info_Complementaire;
          if (AFF_Info_Complementaire) Affichages_complementaires();
          else Ecran_Infos_Complementaires();}
      //======================= Écrans à comportement "standard" ======================
      if (Mode_Ecran < 4 || Mode_Ecran > 6)
         if (Clic_long) if (Augmenter_le_choix) Changer_Heure_Observation();
                        else Changer_le_lieu_Observation();
         else if (Augmenter_le_choix) Changer_Date_Observation();
              else Inverser_test_batterie();}
}//============================= Fin de la boucle de base ====================================

void Erreur_si_BP_central() {
  Teste_activation_BP_central(); if (BP_central_a_traiter) {BIP(); BP_central_a_traiter = false;}}
  
void Inverser_test_batterie() {
  Non_test_accumulateur = !Non_test_accumulateur;
  display.clear(); Affiche_cadre_du_haut(10); display.print("Test Accu : ");
  if (Non_test_accumulateur) display.print("NON"); else display.print("OUI"); display.update();
  delay(1000); Sortie_de_saisie();} 

void Clear_et_Cadre_du_haut(byte posX) {display.clear(); Affiche_cadre_du_haut(posX);} // Pour optimisation.

void Set_curseur_et_AFF_EPROM(byte X, byte Y, int PTR, byte Lgr)  // Pour optimisation du code.
     {display.setCursor(X,Y); Aff_TEXTE_EEPROM(PTR, Lgr);}

void Efface_Zone_et_Set_Curseur(byte Xeff, byte Yeff, byte Lgr, byte posX, byte posY)
    {Effacer_zone(Xeff, Yeff, Lgr); display.setCursor(posX, posY);}   // Pour optimisation du code.

void Ecran_Infos_Complementaires() {
  // Par manque de place mémoire pour le programme le calcul de la journée solaire a été simplifié.
  // Il respecte les valeurs au 16 Septembre et au 22 décembre.
  long Duree_Jour = 86400; int numJ = 0;
  //======================= calcule la durée du jour terrestre =======================
  // Voir https://fr.wikipedia.org/wiki/Dur%C3%A9e_du_jour
  for (byte i = 0; i < Mois; i++) numJ += nbJ_par_mois[i]; numJ = numJ + Jour;
  if (numJ > 356) Duree_Jour = 86425; // Approximation pour la fin de l'année.
  else {if (numJ < 260) Duree_Jour = Duree_Jour + map(numJ, 1,259,20,-21);
       else Duree_Jour = Duree_Jour + map(numJ, 260,356,-21,30);}
  //==================================================================================
  Clear_et_Cadre_du_haut(15); display.print("Donn"); display.print(char(130));
  Aff_TEXTE_EEPROM(926, 12); Set_curseur_et_AFF_EPROM(0,16,823, 21);
  Curseur_Mesure_Affiche(16, 3, Coef_CAN_pour_la_pile);
  Set_curseur_et_AFF_EPROM(0,24,844,12); Aff_TEXTE_EEPROM(835, 9);
  Curseur_Mesure_Affiche(24, 2, Coef_CAN_pour_accumulateur);
  Set_curseur_et_AFF_EPROM(0,33,864,21); Set_curseur_et_AFF_EPROM(0,49,885,41);
  display.setCursor(85,33); display.print(Duree_Jour);
  Set_curseur_et_AFF_EPROM(9,41,778,11);
  // Début calendrier Julien : Premier Janvier 4712 avant J.C.
  display.print(long((4712 + Annee) * 365.24806) + Jour_Num(Mois, Jour));
  display.update();}
  
void Curseur_Mesure_Affiche(byte Y, byte Entree, float Coef_CAN) {
  display.setCursor(84,Y); Mesurer_une_tension(Entree, Coef_CAN); display.print(Tension_lue,1);}
  
void AFF_Tension_lue() {
  CADRE(); display.setTextSize(2); // Taille double.
  display.setCursor(32,25); display.print(Tension_lue,1); display.print("Vcc"); display.setTextSize(1);}

void Fin_alerte_tension_faible() { display.update(); BIP(); ATTENDRE_un_BP();Affiche_Position_Soleil();}

void Teste_la_PILE() {
  Mesurer_une_tension(3, Coef_CAN_pour_la_pile);
  if (Tension_lue < Tension_Pile_minimale) 
     {Clear_et_Cadre_du_haut(9); Aff_TEXTE_EEPROM(544, 19);
      AFF_Tension_lue(); Set_curseur_et_AFF_EPROM(20,47,789, 15); Fin_alerte_tension_faible();}}

void Teste_Accumulateur() {
  if (!Non_test_accumulateur) {
  Mesurer_une_tension(2, Coef_CAN_pour_accumulateur);
  if (Tension_lue < Tension_Accu_minimale) 
     {Clear_et_Cadre_du_haut(30); Aff_TEXTE_EEPROM(844, 12);
      AFF_Tension_lue(); Fin_alerte_tension_faible();}}}      

void Mesurer_une_tension(byte Entre_utilisee, float Coef_de_correction) {
  // Intègre 1000 mesures antiparasites.
  int MAX; Valeurs_lues = 0;
  for (int I = 0; I < (500) ; I++) // Cumuler N mesures. 
      {Valeurs_lues = Valeurs_lues + (analogRead(Entre_utilisee)); }
  Valeurs_lues = (long) Valeurs_lues / 500;
  if (Entre_utilisee == 3) MAX = 5000; else MAX = 9999;
  Tension_lue = map(Valeurs_lues, 0, 1023, 0, MAX); // Tension en volts * 1000.
  Tension_lue = (Tension_lue * Coef_de_correction)/1000;} // Tension en volts corrigée.
  
//******************************** Routines spécifiques aux calculs astronomiques *****************************
int Jour_Num(int M, int J) { // Retourne le jour de l'année pour la Date définie.
  int nbJ = 0; for (byte i = 0; i < M; i++) nbJ += nbJ_par_mois[i]; if (Bissextile) nbJ++;  return nbJ + J;}

// Équation du temps voir les coefficients sur  https://fr.wikipedia.org/wiki/%C3%89quation_du_temps
// Contrairement à ce qui a été indiqué sur un site Internet, la valeur de l'E.T. n'est pas fonction
// de la latitude, elle ne s'annule pas à l'équateur et aux pôles.  
void Calcule_Equation_du_temps(int Ordre_du_jour) {// Valeur retournée en minutes et fractions de minutes.
  float Angle_orbital = 6.283185308 / 365.25 * Ordre_du_jour; 
  Equation_du_Temps = (7.66 * sin(Angle_orbital)) + (9.87 * sin((2 * (Angle_orbital+0.174533))));}
  
float Angle_orbital_Date_heure(int A, int M, int J, int H, int Mn, int S) {
  // Fraction de l'année en radians le jour de l'année "a", au mois "m", au jour "j", 
  // à l'heure "h", la minute "mn" et la seconde "s"
  int Nb = 365; if (Bissextile) Nb = 366;
  return 6.283185308 * (Jour_Num(M, J) - 1 + (H + Mn / 60 + S / 3600 - 12) / 24) / Nb;} // Radians

void eqT_Decl(float Angle_orbital) {
  // Retourne l'équation du temps en minutes et l'angle de déclinaison du soleil en radians
  Equation_du_Temps = (7.66 * sin(Angle_orbital)) + (9.87 * sin((2 * (Angle_orbital + 0.174533))));
  declinaison = 0.006918 - 0.399912 * cos(Angle_orbital) + 0.070257 * sin(Angle_orbital)
                - 0.006758 * cos(2 * Angle_orbital) + 0.000907 * sin(2 * Angle_orbital)
                 - 0.002697 * cos(3 * Angle_orbital) + 0.00148 * sin(3 * Angle_orbital);} // Radians
                
void posSoleil(int A, int M, int J, int H, int Mn, int S, float latitude,
               float longitude, float *Hauteur, float *Azimut) {
  // Latitude et longitude du lieu exprimées en degrés décimaux.
  // La latitude est comprise entre +40° et +55°. (Hémisphère Nord.) (Donc positive dans les calculs.)
  // La longitude est comprise entre -9° et +9°. (Positive si EST et négative si OUest.)
  // L'azimut du soleil est compté positivement dans le sens horairee à partir du Nord.
  eqT_Decl(Angle_orbital_Date_heure(A, M, J, H, Mn, S));
  // Heure légale = Heure solaire vraie + Equation_du_Temps - (4 min par ° EST) + (1H ou 2H si Été ou Hiver) :
  float Correction_pour_Heure_legale = +Equation_du_Temps + (4.0 * longitude) - 120 + (60 * Heure_Hiver);
  float HeureSolaireVraie = H * 60.0 + Mn + (S / 60.0) + Correction_pour_Heure_legale; // En minutes.
  //======================== Ci-dessous : Angle_Solaire_Relatif exprimé en degrés. ========================
  float aSr = (HeureSolaireVraie / 4.0) - 180;
  // Si HSV = 0H aSr = -180 / Si HSV = 12H aSr = 0 / Si HSV = 18H aSr = 90 / Si HSV = 24H aSr = 180 
  latitude = radians(latitude); longitude = radians (longitude);
  float ze = acos(sin(latitude) * sin(declinaison)
             + cos(latitude) * cos(declinaison) * cos(radians(aSr))); // En radians.
  // Ci-dessous, cs est le cosinus de l'azimuth exprimé en radians.
  float cs = -(sin(latitude) * cos(ze) - sin(declinaison)) / (cos(latitude) * sin(ze)); // En radians.
  if (cs < -1)  cs = -1;   // Correction pour palier les imprécisions de calcul.
  else if (cs > 1) cs = 1; // Correction pour palier les imprécisions de calcul.
  *Azimut = degrees(acos(cs)); // Conversion du cosinus  en angle, exprimé en degrés décimaux.
  //================================== Ci-dessous : Calcul de la hauteur ==================================
  if (aSr > 0) *Azimut = 360 - *Azimut; *Hauteur = 90 - (degrees(ze));} // En degrés décimaux.
  
String Conv_Temps_en_chaine(float Temps_en_minutes) {
  // Retourne une chaine au format HH:MM:SS
  // L'entrée est le temps exprimé en minutes et dixièmes de minutes.
  int Hr = Temps_en_minutes / 60;
  int Min = trunc(Temps_en_minutes - 60 * Hr);
  int Sec = trunc((Temps_en_minutes - trunc(Temps_en_minutes)) * 60);
  char stx[10]; // stockage de la chaîne complète
  sprintf(stx, "%02u:%02u:%02u", Hr, Min, Sec); return stx;}  
  
void lmvSoleil(int a, int m, int j, byte Type_de_crepuscule,
               float latitude, float longitude, String *Lever, String *Culmine, String *Coucher) {
  // "EteHiv" vaut 60 en hiver et 120 en Été. (décalage horaire exprimé en minutes.)
  // "Type_de_crepuscule" est le type d'aurore ou de crépuscule retenu :
  //       0 : Crépuscule sur l'horizon avec correction de la réfraction athmosphérique  : r = 0.833°
  //       6 : Crépuscule civil. 12 : Crépuscule nautique. 18 : Crépuscule astronomique.
  eqT_Decl(Angle_orbital_Date_heure(a, m, j, 12, 0, 0));
  latitude = radians(latitude);
  byte EteHiv = 120; if (Heure_Hiver) EteHiv = 60;
  float Acr; // Cosinus de l'Angle crépusculaire.
  switch (Type_de_crepuscule) {
    case 0 : {Acr = cos(radians((90.833 + (0.015 * sqrt(Altitude))))); break;}  // Horizon + correction réfraction.
    case 6 : {Acr = cos(radians(96)); break;} // Aurore et Crépuscules Civils.
    case 12 : {Acr = cos(radians(102)); break;} // Aurore et Crépuscules Nautiques.
    case 18 : Acr = cos(radians(108)); } // Aurore et Crépuscules Astronomiques.
  float aS = acos(Acr / (cos(latitude) * cos(declinaison)) - tan(latitude) * tan(declinaison))
             * Nombre_de_degres_par_Radian;
  Lever_en_min = 720 + (4 * (-longitude - aS)) - Equation_du_Temps + EteHiv;
  Coucher_en_min = 720 + (4 * (-longitude + aS)) - Equation_du_Temps + EteHiv;
  *Lever   = Conv_Temps_en_chaine(Lever_en_min);
  *Coucher = Conv_Temps_en_chaine(Coucher_en_min);
  *Culmine = Conv_Temps_en_chaine(720 + (4 * longitude) - Equation_du_Temps + EteHiv);}
  
void Effectuer_les_calculs() {
  double Longitude_avec_signe;
  Longitude_avec_signe = Longitude; if (!Longitude_EST) Longitude_avec_signe = -Longitude_avec_signe;
  ETE_HIVER(); // Mise à jour heure Été/Hiver.
  Bissextile = (Annee % 4 == 0); // Détermine si bissextile.
  posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude_avec_signe, &Hauteur, &Azimut);
  //-------------------------- Calcule Lever, Méridien, Coucher -------------------------
  lmvSoleil(Annee, Mois, Jour, Mode_Crepuscule, Latitude, Longitude_avec_signe, &Lever, &Culmine, &Coucher);}

//*************************************************************************************************************

void ETE_HIVER() {
  // Détermine le changement d'heure avec une imprecision annuelle de + ou - 3 Jours maximum.
  byte NumJ = Jour_Num(Mois, Jour);
  Heure_Hiver = true;
  if((NumJ > 88) && ((NumJ < 331))) Heure_Hiver = false;}
  
void Affiche_le_nom_de_la_constellation(int PTR, byte Longueur) {
  display.setCursor(25,33);display.print(">>> "); Aff_TEXTE_EEPROM(PTR, Longueur);}
  
void Affiche_constellation_ecliptique() {
  int PTR; byte Longueur;
  if ((Num_Jour_dans_annee < 134) && (Num_Jour_dans_annee > 108)) {PTR = 772; Longueur = 6;} // Bélier
  if ((Num_Jour_dans_annee < 172) && (Num_Jour_dans_annee > 133)) {PTR = 938; Longueur = 7;} // Taureau
  if ((Num_Jour_dans_annee < 201) && (Num_Jour_dans_annee > 171)) {PTR = 945; Longueur = 6;} // Gémaux
  if ((Num_Jour_dans_annee < 222) && (Num_Jour_dans_annee > 200)) {PTR = 951; Longueur = 6;} // Cabcer
  if ((Num_Jour_dans_annee < 259) && (Num_Jour_dans_annee > 221)) {PTR = 957; Longueur = 4;} // Lion
  if ((Num_Jour_dans_annee < 304) && (Num_Jour_dans_annee > 258)) {PTR = 961; Longueur = 6;} // Vierge
  if ((Num_Jour_dans_annee < 327) && (Num_Jour_dans_annee > 303)) {PTR = 967; Longueur = 7;} // Balance
  if ((Num_Jour_dans_annee < 333) && (Num_Jour_dans_annee > 326)) {PTR = 974; Longueur = 8;} // Scorpion
  if ((Num_Jour_dans_annee < 352) && (Num_Jour_dans_annee > 332)) {PTR = 982; Longueur = 8;} // Ophiucus
  if ((Num_Jour_dans_annee > 351) || (Num_Jour_dans_annee < 20)) {PTR = 990; Longueur = 10;} // Sagittaire 
  if ((Num_Jour_dans_annee < 48) && (Num_Jour_dans_annee > 19)) {PTR = 1000; Longueur = 10;} // Capricorne
  if ((Num_Jour_dans_annee < 71) && (Num_Jour_dans_annee > 47)) {PTR = 1010; Longueur = 7;}  // Verseau
  if ((Num_Jour_dans_annee < 109) && (Num_Jour_dans_annee > 70)) {PTR = 571; Longueur = 7;}  // Poisson
  Affiche_le_nom_de_la_constellation(PTR, Longueur); display.update();}
  
void Affichages_complementaires() {
  Clear_et_Cadre_du_haut(12);
  Aff_TEXTE_EEPROM(751, 7); Affiche_la_date(); ETE_HIVER(); 
  CADRE(); Set_curseur_et_AFF_EPROM(23,20,578,8);
  if (Heure_Hiver) Aff_TEXTE_EEPROM(197,6);
  else {display.print(char(130)); display.print("t"); display.print(char(130));}
  display.setCursor(118,20); if (Bissextile) display.print("B");
                             else {display.drawLine(114, 17, 126, 29, WHITE);
                                   display.drawLine(114, 29, 126, 17, WHITE);}
  display.drawLine(113, 17, 113, 29, WHITE); display.drawLine(1, 30, 126, 30, WHITE);
  Affiche_constellation_ecliptique();
  Set_curseur_et_AFF_EPROM(10,43,758, 7);
  display.print(Conv_Temps_en_chaine(Coucher_en_min - Lever_en_min));
  Set_curseur_et_AFF_EPROM(10,53,765, 7);
  display.print(Conv_Temps_en_chaine((1440 - Coucher_en_min) + Lever_en_min));
  display.update();} // 1440 = Nb minutes par jour soit 24 x 60.

void Afficher_coef_deplacement() {
   Efface_Zone_et_Set_Curseur(87, 45, 2,87,45);
   if(Coef_deplacement < 10) display.print("x"); display.print(Coef_deplacement); display.update();}

void Afficher_Jour_valeur_ET_Curseur_Bissextile() {
  // Toutes ces données sont affichées en fonction de la variable [Num_Jour_dans_annee].
  Effacer_les_donnees(); display.setCursor(13,4); // Pour afficher le texte dans le cadre jaune.
  if (Affiche_date) {Aff_TEXTE_EEPROM(856,8); Affiche_la_date();} else Aff_TEXTE_EEPROM(606, 18);
  Bissextile = (Annee % 4 == 0); if (Bissextile) {display.setCursor(40,24); display.print("B");
  display.drawRect(38, 22, 9, 12, WHITE);} else {display.drawRect(38, 22, 9, 12, BLACK);
  Effacer_zone(39, 23, 1);} display.setCursor(89,24); display.print(Num_Jour_dans_annee);
  Calcule_Equation_du_temps(Num_Jour_dans_annee); display.setCursor(40, 51);
  display.print(Equation_du_Temps,1); Colonne = map (Num_Jour_dans_annee,1,365,1,110); Affiche_le_curseur();
  display.update();}

void Efface_le_curseur() {
     display.drawLine(1, 17, 110, 17, BLACK); display.drawLine(1, 18, 110, 18, BLACK);}
     
void Effacer_les_donnees() {
  Effacer_zone(13, 4, 18); // Efface le texte dans le cadre jaune.
  // Effacer les données J, E.T. et Curseur :
  Effacer_zone(89, 24, 3); Effacer_zone(40, 51, 5); Efface_le_curseur();} 

void Trace_Equation_du_temps() {
  byte X, Y; display.clear(); E_T_en_temps_reel = true; Affiche_date = true;
  Rectangle_Graphe(); CADRE_UP();
  // Tracer les graduations sur le coté droit: 
  for (byte I = 19; I < 64; I=I+7) display.drawLine(111, I, 114, I, WHITE);
  // Tracer les valeurs latérales :
  display.setCursor(116,23); display.print("10"); display.setCursor(116,37); display.print("0");
  display.setCursor(116,51); display.print("10"); display.drawLine(1, 40, 110, 40, WHITE);
  display.setCursor(77,24); display.print("J="); display.setCursor(10,51); display.print("E.T.");
  Afficher_coef_deplacement();
  //================== Affiche le graphe de l'équation du temps; ======================
  for (int J = 1; J < 366; J++) {
       Calcule_Equation_du_temps(J);
       X = map (J,1,365,1,110);
       Y = map (Equation_du_Temps,-17,+17,63,18);
       display.drawPixel(X, Y, WHITE);}
  //===================================================================================
  Num_Jour_dans_annee = Jour_Num(Mois, Jour); Afficher_Jour_valeur_ET_Curseur_Bissextile();  
  display.update();}

void Rectangle_Graphe() {display.drawRect(0, 16, 112, 48, WHITE);}

void Initialiser_Horloge_interne(){
  Clear_et_Cadre_du_haut(5); Aff_TEXTE_EEPROM(587, 20); Set_curseur_et_AFF_EPROM(0,16,624,100);
  Aff_TEXTE_EEPROM(724, 36); // Deux instructions car 136 caractères.
  display.update();
  Clic_sur_le_BP = false;
  while (!Clic_sur_le_BP) Tester_le_BP_A0();
  if (Clic_long) {
    rtc.write_protect(false); rtc.halt(false);
    Time t(Annee, Mois, Jour, Heure, Minutes, Secondes, 1); rtc.time(t);} 
  Clic_sur_le_BP = false; Clic_long = false;}    

void Affiche_HORLOGE_temps_reel(){
  byte Temporisation = 15;
  Rotation = false;
  while (!Rotation) {
     display.clear(); CADRE(); display.drawLine(1, 38, 126, 38, WHITE); 
     if (Temporisation > 0) {Temporisation--; Affiche_cadre_du_haut(9); Aff_TEXTE_EEPROM(804, 19);}
     display.setTextSize(2); Time t = rtc.time();       
     display.setCursor(6,20); if (t.date < 10) display.print("0"); display.print(t.date); Affiche_1_espace();
     if (t.mon < 10) display.print("0"); display.print(t.mon); Affiche_1_espace(); display.print(t.yr);
     display.setCursor(18,43); if (t.hr < 10) display.print("0"); display.print(t.hr); Affiche_1_espace();
     if (t.min < 10) display.print("0"); display.print(t.min); Affiche_1_espace();
     if (t.sec < 10) display.print("0"); display.print(t.sec); display.update();
     //===================== Tester le clavier à deux B.P. ==================
     Tester_le_BP_A0();
     if (Clic_sur_le_BP)
        {Clic_sur_le_BP = false;
         if (Clic_long) Initialiser_Horloge_interne(); else BIP();}
     //======================= BP central en sous fonction ==================
     Erreur_si_BP_central(); // Si BP central activé provoque un BIP.
     //========================= Traite la rotation =========================
     INT_active = true; // Armer un nouveau comptage.
     if (Rotation) Clic_long = true;}
  //============================= Sortie du mode GRAPHIQUE ===============================
  display.setTextSize(1);// if (Sens_positif) Mode_Ecran = 0; else Mode_Ecran = 5; 
  Clic_sur_le_BP = false; Affiche_Position_Soleil();}

void Faire_evoluer_le_temps() { // On calcule toutes les deux minutes.
  Minutes = Minutes + 2; if (Minutes > 59) {Minutes = 0; Heure++;}}
  
void Avancer_le_temps() {
  Faire_evoluer_le_temps();
  posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude, &Hauteur, &Azimut);}
 
void Trace_Soleil() {
  byte X,Y;
  savH = Heure; savM = Minutes; savS = Secondes; Heure = 2; Minutes = 0; Secondes = 0;
  //================= Attendre d'être dans la zone affichable ================== 
  while (Azimut < 40) { 
         posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude, &Hauteur, &Azimut);
         Faire_evoluer_le_temps();}
         Avancer_le_temps();
  while (Hauteur < -20) {
        Avancer_le_temps();}
  //================ Tracer les points dans la zone affichable =================
  while ((Hauteur > -20) && (Azimut < 308))
        {Faire_evoluer_le_temps(); posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude, &Hauteur, &Azimut);
         X = map(Azimut,41,308,2,112); Y = map(Hauteur,-20,70,63,18);
         display.drawPixel(X, Y, WHITE);}
  //============================================================================
  Heure = savH; Minutes = savM; Secondes = savS;}

void Affiche_les_reperes(){
  for (byte I=4; I < 112; I=I+4) display.drawLine(I, 59, I, 62, WHITE); 
  display.drawLine(20, 17, 20, 56, WHITE); display.drawLine(56, 17, 56, 56, WHITE);  
  display.drawLine(92, 17, 92, 56, WHITE);
  Efface_Zone_et_Set_Curseur(17, 57, 1,18,57); display.print("E");
  Efface_Zone_et_Set_Curseur(52,57, 1,53,57); display.print("S");
  Efface_Zone_et_Set_Curseur(89, 57, 1,90,57); display.print("W");
  // Tracer les repères horizontaux : 
  for (byte I=23; I < 50; I=I+10) 
       for (byte J=0; J < 109; J=J+5) display.drawLine(J+1, I, J+2, I, WHITE);}

void Efface_zone_date() {Effacer_zone(7,4,10); Effacer_zone(75,4,8);}
  
void Affiche_ecran_Graphique() {
  display.clear(); Mise_a_jour_Date_et_Heure(); Rectangle_Graphe();
  if (Aff_graduations) Affiche_les_reperes();
  // Tracer les graduations sur le coté droit: 
  for (byte I = 18; I < 64; I=I+5) display.drawLine(111, I, 114, I, WHITE);
  // Tracer les valeurs latérales.
  display.setCursor(116,20); display.print("60"); display.setCursor(116,30); display.print("40");  
  display.setCursor(116,40); display.print("20"); display.setCursor(116,50); display.print("0");
  display.drawLine(1, 53, 110, 53, WHITE);
  Efface_zone_date(); Set_curseur_et_AFF_EPROM(7,4,165, 19); display.update();
  if (Mode_trois_courbes)
     {savJ = Jour; savMois = Mois; Mois = 6; Jour = 21; Trace_Soleil(); Mois = 9; Jour = 23; Trace_Soleil();
      Mois = 12; Jour = 21; Trace_Soleil(); Jour = savJ; Mois = savMois;}
  else Trace_Soleil();
  if (Mode_trois_courbes)
       {Effacer_zone(7,4,19); display.drawLine(70, 1, 70, 14, BLACK);
        Set_curseur_et_AFF_EPROM(8,4,125, 19);}
  else {Efface_zone_date(); Mise_a_jour_Date_et_Heure();}
  display.update();}
  
void Affiche_le_curseur() {
  display.drawLine(Colonne-1, 17, Colonne+1, 17, WHITE);
  display.drawPixel(Colonne, 18, WHITE); display.update();}
  
void Affiche_coef_deplacement() {
  Efface_Zone_et_Set_Curseur(96, 20, 2,96,20); display.print("x");
  display.print(OCTET); display.update();}
  
void Determine_parametres_pour_le_curseur() {
  savH = Heure; savM = Minutes; savS = Secondes; Heure = 2; Minutes = 0; Secondes = 0; Azimut = 30;
  Pos_X_sur_ecran = map(Colonne,2,109,43,315); // Pos_X_sur_ecran contient Azimut correspondant à Colonne.
  Efface_Zone_et_Set_Curseur(4, 20, 3,4,20); display.print(Pos_X_sur_ecran); display.update(); 
  //================= Attendre d'être à l'Azimut du curseur ================== 
  while (Azimut < Pos_X_sur_ecran) 
        {posSoleil(Annee, Mois, Jour, Heure, Minutes, Secondes, Latitude, Longitude,
                   &Hauteur, &Azimut); Faire_evoluer_le_temps();}
  //================ Afficher l'heure correspondante =================
  Effacer_zone(75, 4, 8); AFF_HEURE(); display.update();
  //============================================================================
  Heure = savH; Minutes = savM; Secondes = savS;}
  
void Teste_activation_BP_central() {
  if (analogRead(BP_central) < 512) {// Si le B.P. central est activé :
     // ========== Attendre le relacher du B.P. central. ==========
     digitalWrite(LED_Clavier, HIGH); // Allume la LED.
     while (analogRead(BP_central) < 512); // Attendre le relâché.
     digitalWrite(LED_Clavier, LOW);
     // ===========================================================
     BP_central_a_traiter = true;}}

void Traite_BP_Central() {// B.P. central activé dans la boucle de base.
     //============================== Traiter le mode GRAPHIQUE ==============================
     Mode_trois_courbes = false; Aff_graduations = false; Affiche_ecran_Graphique();
     OCTET = 1; // OCTET ici représente le coef de déplacement latéral.
     Colonne = 2; // Positionne le curseur pour le bouton rotatif.
     Affiche_le_curseur(); Affiche_coef_deplacement(); Clic_long = false; 
     while (!Clic_long) {
       //===================== Tester le clavier à deux B.P. ==================
       Tester_le_BP_A0();
       if (Clic_sur_le_BP && !Clic_long) {
           Clic_sur_le_BP= false;
           if (Augmenter_le_choix) {Mode_trois_courbes = !Mode_trois_courbes; Affiche_ecran_Graphique();}
           else {Aff_graduations = !Aff_graduations; Affiche_ecran_Graphique();} }
       //======================= BP central en sous fonction ==================
       Teste_activation_BP_central();
           if (BP_central_a_traiter) {
               BP_central_a_traiter = false; OCTET = OCTET << 1;
               if (OCTET==16) OCTET=1; Affiche_coef_deplacement();}           
       //========================= Traite la rotation =========================
       INT_active = true; // Armer un nouveau comptage.
       if (Rotation) {
          Rotation = false;       
          if (Mode_trois_courbes) BIP();
          else {Efface_le_curseur();
                if (Sens_positif) {Colonne = Colonne + OCTET; if (Colonne > 109) Colonne = 2;}
                else {if (Colonne < OCTET+2) Colonne = 109; else Colonne = Colonne - OCTET;}
                Affiche_le_curseur();
                Determine_parametres_pour_le_curseur();} } }
  //============================= Sortie du mode GRAPHIQUE ===============================
  Clic_sur_le_BP = false;
 // Effectuer_les_calculs(); Affiche_Position_Soleil();
 Mode_Ecran = 2; Temps_reel=true;}

void Modifie_element() {INT_active = true; // Armer un nouveau comptage.
  int MAXIMUM, VALEUR;
  INT_active = true; // Armer un nouveau comptage.
  if (Rotation) { Rotation = false;
     //-------------------------- Positionner le maximum possible --------------------------
     if (Edite_HEURE)
          switch (Mode_Edition)
            {case 0 : {VALEUR = Heure; MAXIMUM = 23; break;}
             case 1 : {VALEUR = Minutes; MAXIMUM = 59; break;}
             case 2 : {VALEUR = Secondes; MAXIMUM = 59;}}
     else switch (Mode_Edition)
            {case 0 : {VALEUR = Jour; MAXIMUM = 31; break;}
             case 1 : {VALEUR = Mois; MAXIMUM = 12; break;}
             default : VALEUR = Annee; MAXIMUM = 3000; break;}  
     //----------------------------- Calcule l'élément pointé ------------------------------
        if (Sens_positif) {VALEUR++; if (VALEUR > MAXIMUM) VALEUR = 0;
            if (VALEUR == 0 && !Edite_HEURE) VALEUR = 1;}
        else {if (VALEUR > 0) VALEUR--; else VALEUR = MAXIMUM;}
        if (VALEUR==0 && !Edite_HEURE) VALEUR = MAXIMUM;
      //------------------------- Mise à jour de l'élément pointé --------------------------
     if (Edite_HEURE)
          switch (Mode_Edition)
            {case 0 : {Heure = VALEUR; break;}
             case 1 : {Minutes = VALEUR; break;}
             default : {Secondes = VALEUR;}}
     else switch (Mode_Edition)
            {case 0 : {Jour = VALEUR; break;}
             case 1 : {Mois = VALEUR; break;}
             default : Annee = VALEUR; }
     SAVA = Annee; SAVM = Mois; SAVJ = Jour; SAVH = Heure; SAVmin = Minutes, SAVs = Secondes;    
     Effacer_zone(7, 4, 19); Mise_a_jour_Date_et_Heure(); display.update();}}
     
void Changer_Date_Observation() {
  byte Date_speciale = 0;
  Prepare_pour_DATE_et_HEURE(); Edite_HEURE = false; // Donc on édite la date.
  Aff_TEXTE_EEPROM(236, 16);
  Set_curseur_et_AFF_EPROM(5,36,298, 19); Positionne_curseur_DATE();
  while (!Clic_long) {
    Tester_le_BP_A0();
    if (Clic_sur_le_BP && !Clic_long) {
         Clic_sur_le_BP= false;
         if (Augmenter_le_choix) {
            Clic_sur_le_BP= false;  Mode_Edition++; if (Mode_Edition == 3) Mode_Edition = 0;
            Positionne_curseur_DATE();}
         else {savJ = Jour; savMois = Mois;
              Efface_Zone_et_Set_Curseur(14, 48, 17,16,48); // Efface le pointeur ou texte.
                                              // Pour afficher la nature de la date sélectionnée. 
              switch (Date_speciale)
                {case 0 : {Restituer_Simulation(); Secondes = SAVs; // > Jour simulé < 
                           Aff_TEXTE_EEPROM(220, 16); break;} 
                 case 1 : {Mois = 3; Jour = 21; Aff_TEXTE_EEPROM(148, 17); break;} // Equinoxe Printmps
                 case 2 : {Mois = 6; Jour = 21; Aff_TEXTE_EEPROM(184, 13); break;} // Solstice Eté.
                 case 3 : {Mois = 9; Jour = 23; Aff_TEXTE_EEPROM(203, 16); break;} // Equinoxe Automne.
                 case 4 : {Mois = 12; Jour = 21; Aff_TEXTE_EEPROM(125, 8);
                           Affiche_1_espace(); Aff_TEXTE_EEPROM(197, 6);} } // Solstice Hiver.          
             Effacer_zone(7, 4, 19); Mise_a_jour_Date_et_Heure(); display.update();
             Date_speciale++; if (Date_speciale == 5) Date_speciale = 0;} }        
    Modifie_element();} //Si Bouton rotation modifie la valeur et affiche les changements.
  Sortie_de_saisie();}
  
void Restituer_Simulation() {
  Annee = SAVA; Mois = SAVM; Jour = SAVJ; Heure = SAVH; Minutes = SAVmin; Secondes = SAVs;}
  
void Transfere_Donnees_DS1302() {
  Time t = rtc.time();
  Annee = t.yr; Mois = t.mon; Jour = t.date; Heure = t.hr; Minutes = t.min; Secondes = t.sec;}
  
void Prepare_pour_DATE_et_HEURE() {
  Mode_Edition = 0; Restituer_Simulation(); Restituer_Simulation(); Affiche_date_et_heure();
  Prepare_cadre_3_donnees(); Clic_long = false;}
  
void Changer_Heure_Observation() {
  Edite_HEURE = true; // Donc on édite la date.
  Prepare_pour_DATE_et_HEURE(); Aff_TEXTE_EEPROM(272, 16); Positionne_curseur_DATE();
  while (!Clic_long) {
    Tester_le_BP_A0();
    if (Clic_sur_le_BP && !Clic_long) {
         Clic_sur_le_BP= false;  Mode_Edition++; if (Mode_Edition == 3) Mode_Edition = 0;
         Positionne_curseur_DATE();}
    Modifie_element(); } //Si Bouton rotation modifie la valeur et affiche les changements.
  Sortie_de_saisie();}

void Positionne_curseur_DATE() {
  Efface_Zone_et_Set_Curseur(14, 48, 17,16,48); // Efface le pointeur ou texte.
  if (Mode_Edition ==1) display.setCursor(54,48);
  if (Mode_Edition ==2) display.setCursor(100,48); display.print("^"); display.update();}
  
void Positionne_curseur_ALTITUDE() {
  Efface_Zone_et_Set_Curseur(12, 48, 19,16,48); // Efface le pointeur. 
  switch (Mode_Edition)
         {case 0 : {display.setCursor(14,48); break;}
          case 1 : {display.setCursor(53,48); break;}
          case 2 : {display.setCursor(92,48); break;}
          default : display.setCursor(119,48);}
  display.print("^"); display.update();}
  
void Rafraichi_Altitude() {
  Efface_Zone_et_Set_Curseur(50, 4, 6, 50,4); display.print(Altitude);
  display.print(" m"); display.update();} 
    
void Rafraichir_type() {// procedure locale
    Efface_Zone_et_Set_Curseur(65, 20, 9,65,20);
    if (Edit_Longitude) Aff_TEXTE_EEPROM(252, 9); else Aff_TEXTE_EEPROM(288, 8); display.update();}
    
void AFF_E_W() {if (Longitude_EST) display.print(" E"); else display.print(" W");}

void Rafraichir_Position() {
  Efface_Zone_et_Set_Curseur(31, 4, 11,31,4); if (Edit_Longitude) Affiche_1_espace(); 
  if (Edit_Longitude) {display.print(Longitude,5); AFF_E_W();}
  else {display.print(Latitude,5); display.print(" N");} }

void Positionne_curseur_Coordonnees() {
  Efface_Zone_et_Set_Curseur(37, 48, 7, 37 + (Mode_Edition * 6),48);  // Efface le pointeur.
  display.print("^"); display.update();}

void Changer_le_lieu_Observation() {
  //=================================== Changer l'ALTITUDE ======================================
  Mode_Edition = 0; Edite_HEURE = true; // Donc on édite la date.
  display.clear(); Rafraichi_Altitude();
  Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(325,8); Aff_TEXTE_EEPROM(339,9);
  Set_curseur_et_AFF_EPROM(5,36,348,20); Positionne_curseur_ALTITUDE();
  Clic_long = false;
  while (!Clic_long) {
    //------------------------- Traite le B.P. central -----------------------
    Erreur_si_BP_central();
    //--------------------------- Traite le clavier --------------------------
    Tester_le_BP_A0();
    if (Clic_sur_le_BP && !Clic_long) {
         Clic_sur_le_BP= false;  Mode_Edition++; if (Mode_Edition == 4) Mode_Edition = 0;
         Positionne_curseur_ALTITUDE();}
    //-------------------------- Traite la rotation --------------------------
    INT_active = true; // Armer un nouveau comptage.
    if (Rotation) { Rotation = false;
       if (Sens_positif) {        
       switch (Mode_Edition) // Maximum 9999m limité par l'affichage de la page des données d'observation.
              {case 0 : {Altitude = Altitude + 1000; break;}
               case 1 : {Altitude = Altitude + 100; break;}
               case 2 : {Altitude = Altitude + 10; break;}
              default : Altitude++;}; if (Altitude > 9999){ Altitude = 9999; BIP();}}
     else{switch (Mode_Edition) // Minimum zéro limité par l'affichage de la page des données.
              {case 0 : {Altitude = Altitude - 1000; break;}
               case 1 : {Altitude = Altitude - 100; break;}
               case 2 : {Altitude = Altitude - 10; break;}
              default : Altitude--;}; if (Altitude < 0){ Altitude = 0; BIP();}}       
       Rafraichi_Altitude();}
     //-----------------------------------------------------------------------
  Clic_sur_le_BP= false;}
  //============================== Changer les COORDONNEES GPS ==================================
  Edit_Longitude = true; Mode_Edition = 0; 
  display.clear(); Rafraichir_Position();
  Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(324, 9); Rafraichir_type();
  Set_curseur_et_AFF_EPROM(31,36,367, 8); Positionne_curseur_Coordonnees();
  Clic_long = false;
  while (!Clic_long) {
    //------------------------- Traite le B.P. central -----------------------
    Erreur_si_BP_central();
    //--------------------------- Traite le clavier --------------------------
    Tester_le_BP_A0();
    if (Clic_sur_le_BP && Clic_long && !Augmenter_le_choix) {
        Clic_long = false; Clic_sur_le_BP = false;
        if (Edit_Longitude) { Longitude_EST = !Longitude_EST; Rafraichir_Position(); display.update();} else BIP();} 
    if (Clic_sur_le_BP && !Clic_long) {
         Clic_sur_le_BP= false;
         if (Augmenter_le_choix) 
            {Mode_Edition++;
             if (Mode_Edition == 7) Mode_Edition = 0; if (Mode_Edition == 1) Mode_Edition = 2;
         Positionne_curseur_Coordonnees();}
         else {Edit_Longitude = !Edit_Longitude; Rafraichir_Position(); Rafraichir_type();
               Mode_Edition = 0; Positionne_curseur_Coordonnees();}}
    //-------------------------- Traite la rotation --------------------------
    INT_active = true; // Armer un nouveau comptage.
    if (Rotation) { Rotation = false;
       if (Edit_Longitude) VALEUR = Longitude; else VALEUR = Latitude;
       if (Sens_positif) {        
       switch (Mode_Edition)
              {case 0 : {VALEUR++; break;}
               case 2 : {VALEUR = VALEUR + 0.1; break;}
               case 3 : {VALEUR = VALEUR + 0.01; break;}
               case 4 : {VALEUR = VALEUR + 0.001; break;}
               case 5 : {VALEUR = VALEUR + 0.0001; break;}
               case 6 :  VALEUR = VALEUR + 0.00001;} // 6 obligé car 1 ignoré. (Point décimal)
               //----- Traiter la limite supérieure -----
               if (Edit_Longitude && VALEUR > 9) {VALEUR = 9; BIP();}
               if (!Edit_Longitude && VALEUR > 55) {VALEUR = 55; BIP();} }
       else {switch (Mode_Edition)
              {case 0 : {VALEUR--; break;}
               case 2 : {VALEUR = VALEUR - 0.1; break;}
               case 3 : {VALEUR = VALEUR - 0.01; break;}
               case 4 : {VALEUR = VALEUR - 0.001; break;}
               case 5 : {VALEUR = VALEUR - 0.0001; break;}
               case 6 :  VALEUR = VALEUR - 0.00001;} // 6 obligé car 1 ignoré. (Point décimal)
               //----- Traiter la limite inférieure -----
               if (Edit_Longitude && VALEUR < 0) {VALEUR = 0; BIP();}
               if (!Edit_Longitude && VALEUR < 40) {VALEUR = 40; BIP();}}
     if (Edit_Longitude) Longitude = VALEUR; else Latitude = VALEUR;  
     Rafraichir_Position(); display.update();}}
     //-----------------------------------------------------------------------  
  Sortie_de_saisie();}  
  //=============================================================================================
  
void Sortie_de_saisie() {
  Clic_sur_le_BP= false; Effectuer_les_calculs(); Affiche_Position_Soleil(); Mode_Ecran = 2; Temps_reel=true;}
  
void Changer_mode_crepusculaire() {
  Clic_long = false; Affiche_mode_crepusculaire();
  while (!Clic_long) {
        //=================================================================
        if (Rotation) {Rotation = false; Clic_long = true;
            if (Sens_positif) {Mode_Ecran = 7; Affiche_la_Position();}
            else {Mode_Ecran = 5; Affichages_complementaires();}}
        //=================================================================  
        Tester_le_BP_A0();
        if (Clic_sur_le_BP && !Clic_long)
           {Clic_sur_le_BP = false; 
            Mode_Crepuscule = Mode_Crepuscule + 6; if (Mode_Crepuscule == 24) Mode_Crepuscule = 0;
            Affiche_mode_crepusculaire();}
        //=================================================================
        Erreur_si_BP_central();}       
        //=================================================================
 if (Mode_Ecran != 5 && Mode_Ecran != 7) Sortie_de_saisie();}

void Affiche_mode_crepusculaire() {
  Clear_et_Cadre_du_haut(12); Aff_TEXTE_EEPROM(375, 18);
  CADRE(); Set_curseur_et_AFF_EPROM(7,20,393,19);
  display.drawLine(1, 30, 126, 30, WHITE); display.setCursor(5,42); 
  if (Mode_Crepuscule==0) Set_curseur_et_AFF_EPROM(5,38,317, 7);
  switch (Mode_Crepuscule) {
        case  0 : {Aff_TEXTE_EEPROM(262, 9); Set_curseur_et_AFF_EPROM(5,48,412,20); break;} 
        case  6 : {Aff_TEXTE_EEPROM(433, 6); break;}
        case 12 : {Aff_TEXTE_EEPROM(439, 9); break;}
        case 18 : Aff_TEXTE_EEPROM(448, 13);}
  display.update(); }

void Efface_ecran() {
  display.clear(); display.setTextSize(2); Set_curseur_et_AFF_EPROM(15,20,563, 8);
  display.update(); delay(1000); display.clear(); display.update(); display.setTextSize(1);}

//-------------------- Routines de servitude pour traiter le capteur incrémental -------------------
void Delay_antirebonds() {for(int I=0; I<100; I++);} // Valeur de 3 expérimentale.

void Front_sur_A(){ // Interruption quand A change d'état avant B.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
  if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
    A_set = !A_set;
    if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }

void Front_sur_B(){ // Interruption quand B change d'état avant A.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
  if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
    B_set = !B_set;
    // Décrémenter le compteur si B avant A.
    if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
//--------------------------------------------------------------------------------------------------
  
void Affiche_la_Position() {
  display.clear(); Affiche_cadre_du_haut(9); display.print("Lieu d'observation.");
  Prepare_cadre_3_donnees(); display.drawLine(1, 30, 62, 30, BLACK);
  Aff_TEXTE_EEPROM(461, 14); display.print(Altitude); display.print("m");
  display.drawLine(63, 17, 63, 30, WHITE); Set_curseur_et_AFF_EPROM(10,36,288, 9);
  display.print(Latitude,5); display.print(" N");
  Set_curseur_et_AFF_EPROM(10,48,252, 10); display.print(Longitude,5);AFF_E_W(); 
  display.update();} 

void Affiche_Lever_Coucher() {
  Affiche_date_et_heure();
  display.drawLine(71, 1, 126, 14, WHITE); display.drawLine(71, 14, 126, 1, WHITE); // Barre l'heure.
  Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(475,10); display.print(Culmine);
  Set_curseur_et_AFF_EPROM(10,36,524,10); display.print(Lever); Set_curseur_et_AFF_EPROM(10,48,534,10);
  display.print(Coucher); display.update();} 

void Affiche_Position_Soleil() {
  Affiche_date_et_heure(); Prepare_cadre_3_donnees(); Aff_TEXTE_EEPROM(485, 19);
  Set_curseur_et_AFF_EPROM(15,36,504, 10); display.print(Hauteur, 3);
  Set_curseur_et_AFF_EPROM(15,48,514, 10); display.print(Azimut, 3); display.update();}
 
void Prepare_cadre_3_donnees() {
  CADRE(); display.drawLine(1, 30, 126, 30, WHITE); display.setCursor(9,20); } 

void AFF_HEURE() {display.setCursor(75,4); Complete_a_10(Heure); display.print(Heure); Affiche_separation();
  Complete_a_10(Minutes); display.print(Minutes); Affiche_separation(); Complete_a_10(Secondes);
  display.print(Secondes);}
  
void Affiche_la_date() {
  Complete_a_10(Jour); display.print(Jour); Affiche_1_espace(); Complete_a_10(Mois); display.print(Mois);
  Affiche_1_espace(); display.print(Annee);} 
  
void Mise_a_jour_Date_et_Heure()
  {Affiche_cadre_du_haut(7); Affiche_la_date(); AFF_HEURE(); display.drawLine(70, 1, 70, 14, WHITE);}

void Affiche_date_et_heure() {display.clear(); Mise_a_jour_Date_et_Heure();}

void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {
  for (byte I=0; I<Longueur; I++) {display.print(char(EEPROM.read(PTR))); PTR++;}; }

void Complete_a_10(int Valeur) {if(Valeur < 10) display.print("0");}
void Affiche_separation() {display.print(":");}
void Affiche_1_espace() {display.print(" ");}

void Effacer_zone(byte X, byte Y, byte NB_caracteres) {
  byte I = 10; 
  for (byte J=0; J < I; J++)
       display.drawLine(X, Y + J, X + (NB_caracteres * 6), Y + J, BLACK);}
  
void Affiche_cadre_du_haut(byte X) {
  CADRE_UP();  display.setTextSize(1); // Taille la plus petite.
  display.setCursor(X,4);}

void CADRE_UP() {display.drawRect(0, 0, 128, 16, WHITE);}
void CADRE() {display.drawRect(0, 16, 128, 48, WHITE);}

void BIP() {
  pinMode(BUZZER, OUTPUT); tone(BUZZER,3000); delay(100); noTone(BUZZER); pinMode(BUZZER, INPUT);}
  
void ATTENDRE_un_BP() { Clic_sur_le_BP = false;
  while (!Clic_sur_le_BP) {
    digitalWrite(LED_Clavier,!digitalRead(LED_Clavier)); // Inverser la LED clavier.
    delay(50); Tester_le_BP_A0();}
  digitalWrite(LED_Clavier,LOW); Clic_sur_le_BP = false; }
  
// >>>>>>>>>>>>>>> Le B.P. est câblé en "Analogique" <<<<<<<<<<<<<<<<<<
void Tester_le_BP_A0() { // La lecture sur A0 donne 0 ou 930 maximum.
  int Numerisation_BP; // Valeur retournée par la CAN de A0.
  boolean Sauve_etat_LED_jaune;
  byte COMPTEUR; unsigned long Horloge_actuelle;
  Sauve_etat_LED_jaune = digitalRead(LED_Jaune); Numerisation_BP = analogRead(0);
  if ((Numerisation_BP > 710) || (Numerisation_BP < 240)) {
    Clic_sur_le_BP = true; digitalWrite(LED_Clavier, HIGH);
    Augmenter_le_choix = false;
    Horloge_actuelle = millis(); Clic_long = false;
    if (Numerisation_BP > 710) { Augmenter_le_choix = true; }
    delay(100); // Attendre_la_stabilisation du travail.
    // Attendre_la_stabilisation du repos :
    COMPTEUR = 0;
    while (COMPTEUR < NB_tests_antirebonds) {
      if ((millis() - Horloge_actuelle) > 700) {Clic_long = true; digitalWrite(LED_Jaune, HIGH);}
      COMPTEUR++; Numerisation_BP = analogRead(0);
      if ((Numerisation_BP > 710) || (Numerisation_BP < 240)) COMPTEUR = 0; }
  digitalWrite(LED_Clavier, LOW);} // Éteindre la LED clavier.
  digitalWrite(LED_Jaune, Sauve_etat_LED_jaune); } // Restitue l'état de la LED jaune.
  
